# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for commonly used utilities."""

import functools
import inspect
import logging
import urllib

logger = logging.getLogger(__name__)

POSITIONAL_WARNING = "WARNING"
POSITIONAL_EXCEPTION = "EXCEPTION"
POSITIONAL_IGNORE = "IGNORE"
POSITIONAL_SET = frozenset(
    [POSITIONAL_WARNING, POSITIONAL_EXCEPTION, POSITIONAL_IGNORE]
)

positional_parameters_enforcement = POSITIONAL_WARNING

_SYM_LINK_MESSAGE = "File: {0}: Is a symbolic link."
_IS_DIR_MESSAGE = "{0}: Is a directory"
_MISSING_FILE_MESSAGE = "Cannot access {0}: No such file or directory"


def positional(max_positional_args):
    """A decorator to declare that only the first N arguments may be positional.

    This decorator makes it easy to support Python 3 style keyword-only
    parameters. For example, in Python 3 it is possible to write::

        def fn(pos1, *, kwonly1=None, kwonly2=None):
            ...

    All named parameters after ``*`` must be a keyword::

        fn(10, 'kw1', 'kw2')  # Raises exception.
        fn(10, kwonly1='kw1')  # Ok.

    Example
    ^^^^^^^

    To define a function like above, do::

        @positional(1)
        def fn(pos1, kwonly1=None, kwonly2=None):
            ...

    If no default value is provided to a keyword argument, it becomes a
    required keyword argument::

        @positional(0)
        def fn(required_kw):
            ...

    This must be called with the keyword parameter::

        fn()  # Raises exception.
        fn(10)  # Raises exception.
        fn(required_kw=10)  # Ok.

    When defining instance or class methods always remember to account for
    ``self`` and ``cls``::

        class MyClass(object):

            @positional(2)
            def my_method(self, pos1, kwonly1=None):
                ...

            @classmethod
            @positional(2)
            def my_method(cls, pos1, kwonly1=None):
                ...

    The positional decorator behavior is controlled by
    ``_helpers.positional_parameters_enforcement``, which may be set to
    ``POSITIONAL_EXCEPTION``, ``POSITIONAL_WARNING`` or
    ``POSITIONAL_IGNORE`` to raise an exception, log a warning, or do
    nothing, respectively, if a declaration is violated.

    Args:
        max_positional_arguments: Maximum number of positional arguments. All
                                  parameters after this index must be
                                  keyword only.

    Returns:
        A decorator that prevents using arguments after max_positional_args
        from being used as positional parameters.

    Raises:
        TypeError: if a keyword-only argument is provided as a positional
                   parameter, but only if
                   _helpers.positional_parameters_enforcement is set to
                   POSITIONAL_EXCEPTION.
    """

    def positional_decorator(wrapped):
        @functools.wraps(wrapped)
        def positional_wrapper(*args, **kwargs):
            if len(args) > max_positional_args:
                plural_s = ""
                if max_positional_args != 1:
                    plural_s = "s"
                message = (
                    "{function}() takes at most {args_max} positional "
                    "argument{plural} ({args_given} given)".format(
                        function=wrapped.__name__,
                        args_max=max_positional_args,
                        args_given=len(args),
                        plural=plural_s,
                    )
                )
                if positional_parameters_enforcement == POSITIONAL_EXCEPTION:
                    raise TypeError(message)
                elif positional_parameters_enforcement == POSITIONAL_WARNING:
                    logger.warning(message)
            return wrapped(*args, **kwargs)

        return positional_wrapper

    if isinstance(max_positional_args, int):
        return positional_decorator
    else:
        args, _, _, defaults, _, _, _ = inspect.getfullargspec(max_positional_args)
        return positional(len(args) - len(defaults))(max_positional_args)


def parse_unique_urlencoded(content):
    """Parses unique key-value parameters from urlencoded content.

    Args:
        content: string, URL-encoded key-value pairs.

    Returns:
        dict, The key-value pairs from ``content``.

    Raises:
        ValueError: if one of the keys is repeated.
    """
    urlencoded_params = urllib.parse.parse_qs(content)
    params = {}
    for key, value in urlencoded_params.items():
        if len(value) != 1:
            msg = "URL-encoded content contains a repeated value:" "%s -> %s" % (
                key,
                ", ".join(value),
            )
            raise ValueError(msg)
        params[key] = value[0]
    return params


def update_query_params(uri, params):
    """Updates a URI with new query parameters.

    If a given key from ``params`` is repeated in the ``uri``, then
    the URI will be considered invalid and an error will occur.

    If the URI is valid, then each value from ``params`` will
    replace the corresponding value in the query parameters (if
    it exists).

    Args:
        uri: string, A valid URI, with potential existing query parameters.
        params: dict, A dictionary of query parameters.

    Returns:
        The same URI but with the new query parameters added.
    """
    parts = urllib.parse.urlparse(uri)
    query_params = parse_unique_urlencoded(parts.query)
    query_params.update(params)
    new_query = urllib.parse.urlencode(query_params)
    new_parts = parts._replace(query=new_query)
    return urllib.parse.urlunparse(new_parts)


def _add_query_parameter(url, name, value):
    """Adds a query parameter to a url.

    Replaces the current value if it already exists in the URL.

    Args:
        url: string, url to add the query parameter to.
        name: string, query parameter name.
        value: string, query parameter value.

    Returns:
        Updated query parameter. Does not update the url if value is None.
    """
    if value is None:
        return url
    else:
        return update_query_params(url, {name: value})



#!/usr/bin/env python
#
#    Copyright (C) 2009 Google Inc.
#
#   Licensed under the Apache License 2.0;
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


# This module is used for version 2 of the Google Data APIs.
# TODO: add proxy handling.


# __author__ = 'j.s@google.com (Jeff Scudder)'

import http.client
import io
import os
import urllib.error
import urllib.parse
import urllib.request

ssl = None
try:
    import ssl
except ImportError:
    pass


class Error(Exception):
    pass


class UnknownSize(Error):
    pass


class ProxyError(Error):
    pass


MIME_BOUNDARY = 'END_OF_PART'


def get_headers(http_response):
    """Retrieves all HTTP headers from an HTTP response from the server.

    This method is provided for backwards compatibility for Python2.2 and 2.3.
    The httplib.HTTPResponse object in 2.2 and 2.3 does not have a getheaders
    method so this function will use getheaders if available, but if not it
    will retrieve a few using getheader.
    """
    if hasattr(http_response, 'getheaders'):
        return http_response.getheaders()
    else:
        headers = []
        for header in (
                'location', 'content-type', 'content-length', 'age', 'allow',
                'cache-control', 'content-location', 'content-encoding', 'date',
                'etag', 'expires', 'last-modified', 'pragma', 'server',
                'set-cookie', 'transfer-encoding', 'vary', 'via', 'warning',
                'www-authenticate', 'gdata-version'):
            value = http_response.getheader(header, None)
            if value is not None:
                headers.append((header, value))
        return headers


class HttpRequest(object):
    """Contains all of the parameters for an HTTP 1.1 request.

    The HTTP headers are represented by a dictionary, and it is the
    responsibility of the user to ensure that duplicate field names are combined
    into one header value according to the rules in section 4.2 of RFC 2616.
    """
    method = None
    uri = None

    def __init__(self, uri=None, method=None, headers=None):
        """Construct an HTTP request.

        Args:
          uri: The full path or partial path as a Uri object or a string.
          method: The HTTP method for the request, examples include 'GET', 'POST',
                  etc.
          headers: dict of strings The HTTP headers to include in the request.
        """
        self.headers = headers or {}
        self._body_parts = []
        if method is not None:
            self.method = method
        if isinstance(uri, str):
            uri = Uri.parse_uri(uri)
        self.uri = uri or Uri()

    def add_body_part(self, data, mime_type, size=None):
        """Adds data to the HTTP request body.

        If more than one part is added, this is assumed to be a mime-multipart
        request. This method is designed to create MIME 1.0 requests as specified
        in RFC 1341.

        Args:
          data: str or a file-like object containing a part of the request body.
          mime_type: str The MIME type describing the data
          size: int Required if the data is a file like object. If the data is a
                string, the size is calculated so this parameter is ignored.
        """
        if hasattr(data, '__len__'):
            size = len(data)
        if size is None:
            # TODO: support chunked transfer if some of the body is of unknown size.
            raise UnknownSize('Each part of the body must have a known size.')
        if 'Content-Length' in self.headers:
            content_length = int(self.headers['Content-Length'])
        else:
            content_length = 0
        # If this is the first part added to the body, then this is not a multipart
        # request.
        if len(self._body_parts) == 0:
            self.headers['Content-Type'] = mime_type
            content_length = size
            self._body_parts.append(data)
        elif len(self._body_parts) == 1:
            # This is the first member in a mime-multipart request, so change the
            # _body_parts list to indicate a multipart payload.
            self._body_parts.insert(0, 'Media multipart posting')
            boundary_string = '\r\n--%s\r\n' % (MIME_BOUNDARY,)
            content_length += len(boundary_string) + size
            self._body_parts.insert(1, boundary_string)
            content_length += len('Media multipart posting')
            # Put the content type of the first part of the body into the multipart
            # payload.
            original_type_string = 'Content-Type: %s\r\n\r\n' % (
                self.headers['Content-Type'],)
            self._body_parts.insert(2, original_type_string)
            content_length += len(original_type_string)
            boundary_string = '\r\n--%s\r\n' % (MIME_BOUNDARY,)
            self._body_parts.append(boundary_string)
            content_length += len(boundary_string)
            # Change the headers to indicate this is now a mime multipart request.
            self.headers['Content-Type'] = 'multipart/related; boundary="%s"' % (
                MIME_BOUNDARY,)
            self.headers['MIME-version'] = '1.0'
            # Include the mime type of this part.
            type_string = 'Content-Type: %s\r\n\r\n' % (mime_type)
            self._body_parts.append(type_string)
            content_length += len(type_string)
            self._body_parts.append(data)
            ending_boundary_string = '\r\n--%s--' % (MIME_BOUNDARY,)
            self._body_parts.append(ending_boundary_string)
            content_length += len(ending_boundary_string)
        else:
            # This is a mime multipart request.
            boundary_string = '\r\n--%s\r\n' % (MIME_BOUNDARY,)
            self._body_parts.insert(-1, boundary_string)
            content_length += len(boundary_string) + size
            # Include the mime type of this part.
            type_string = 'Content-Type: %s\r\n\r\n' % (mime_type)
            self._body_parts.insert(-1, type_string)
            content_length += len(type_string)
            self._body_parts.insert(-1, data)
        self.headers['Content-Length'] = str(content_length)

    # I could add an "append_to_body_part" method as well.

    AddBodyPart = add_body_part

    def add_form_inputs(self, form_data,
                        mime_type='application/x-www-form-urlencoded'):
        """Form-encodes and adds data to the request body.

        Args:
          form_data: dict or sequnce or two member tuples which contains the
                     form keys and values.
          mime_type: str The MIME type of the form data being sent. Defaults
                     to 'application/x-www-form-urlencoded'.
        """
        body = urllib.parse.urlencode(form_data)
        self.add_body_part(body, bytes(mime_type, 'ascii'))

    AddFormInputs = add_form_inputs

    def _copy(self):
        """Creates a deep copy of this request."""
        copied_uri = Uri(self.uri.scheme, self.uri.host, self.uri.port,
                         self.uri.path, self.uri.query.copy())
        new_request = HttpRequest(uri=copied_uri, method=self.method,
                                  headers=self.headers.copy())
        new_request._body_parts = self._body_parts[:]
        return new_request

    def _dump(self):
        """Converts to a printable string for debugging purposes.

        In order to preserve the request, it does not read from file-like objects
        in the body.
        """
        output = 'HTTP Request\n  method: %s\n  url: %s\n  headers:\n' % (
            self.method, str(self.uri))
        for header, value in self.headers.items():
            output += '    %s: %s\n' % (header, value)
        output += '  body sections:\n'
        i = 0
        for part in self._body_parts:
            if isinstance(part, str):
                output += '    %s: %s\n' % (i, part)
            else:
                output += '    %s: <file like object>\n' % i
            i += 1
        return output


def _apply_defaults(http_request):
    if http_request.uri.scheme is None:
        if http_request.uri.port == 443:
            http_request.uri.scheme = 'https'
        else:
            http_request.uri.scheme = 'http'


class Uri(object):
    """A URI as used in HTTP 1.1"""
    scheme = None
    host = None
    port = None
    path = None

    def __init__(self, scheme=None, host=None, port=None, path=None, query=None):
        """Constructor for a URI.

        Args:
          scheme: str This is usually 'http' or 'https'.
          host: str The host name or IP address of the desired server.
          post: int The server's port number.
          path: str The path of the resource following the host. This begins with
                a /, example: '/calendar/feeds/default/allcalendars/full'
          query: dict of strings The URL query parameters. The keys and values are
                 both escaped so this dict should contain the unescaped values.
                 For example {'my key': 'val', 'second': '!!!'} will become
                 '?my+key=val&second=%21%21%21' which is appended to the path.
        """
        self.query = query or {}
        if scheme is not None:
            self.scheme = scheme
        if host is not None:
            self.host = host
        if port is not None:
            self.port = port
        if path:
            self.path = path

    def _get_query_string(self):
        param_pairs = []
        for key, value in self.query.items():
            quoted_key = urllib.parse.quote_plus(str(key))
            if value is None:
                param_pairs.append(quoted_key)
            else:
                quoted_value = urllib.parse.quote_plus(str(value))
                param_pairs.append('%s=%s' % (quoted_key, quoted_value))
        return '&'.join(param_pairs)

    def _get_relative_path(self):
        """Returns the path with the query parameters escaped and appended."""
        param_string = self._get_query_string()
        if self.path is None:
            path = '/'
        else:
            path = self.path
        if param_string:
            return '?'.join([path, param_string])
        else:
            return path

    def _to_string(self):
        if self.scheme is None and self.port == 443:
            scheme = 'https'
        elif self.scheme is None:
            scheme = 'http'
        else:
            scheme = self.scheme
        if self.path is None:
            path = '/'
        else:
            path = self.path
        if self.port is None:
            return '%s://%s%s' % (scheme, self.host, self._get_relative_path())
        else:
            return '%s://%s:%s%s' % (scheme, self.host, str(self.port),
                                     self._get_relative_path())

    def __str__(self):
        return self._to_string()

    def modify_request(self, http_request=None):
        """Sets HTTP request components based on the URI."""
        if http_request is None:
            http_request = HttpRequest()
        if http_request.uri is None:
            http_request.uri = Uri()
        # Determine the correct scheme.
        if self.scheme:
            http_request.uri.scheme = self.scheme
        if self.port:
            http_request.uri.port = self.port
        if self.host:
            http_request.uri.host = self.host
        # Set the relative uri path
        if self.path:
            http_request.uri.path = self.path
        if self.query:
            http_request.uri.query = self.query.copy()
        return http_request

    ModifyRequest = modify_request

    def parse_uri(uri_string):
        """Creates a Uri object which corresponds to the URI string.

        This method can accept partial URIs, but it will leave missing
        members of the Uri unset.
        """
        parts = urllib.parse.urlparse(uri_string)
        uri = Uri()
        if parts[0]:
            uri.scheme = parts[0]
        if parts[1]:
            host_parts = parts[1].split(':')
            if host_parts[0]:
                uri.host = host_parts[0]
            if len(host_parts) > 1:
                uri.port = int(host_parts[1])
        if parts[2]:
            uri.path = parts[2]
        if parts[4]:
            param_pairs = parts[4].split('&')
            for pair in param_pairs:
                pair_parts = pair.split('=')
                if len(pair_parts) > 1:
                    uri.query[urllib.parse.unquote_plus(pair_parts[0])] = (
                        urllib.parse.unquote_plus(pair_parts[1]))
                elif len(pair_parts) == 1:
                    uri.query[urllib.parse.unquote_plus(pair_parts[0])] = None
        return uri

    parse_uri = staticmethod(parse_uri)

    ParseUri = parse_uri


parse_uri = Uri.parse_uri

ParseUri = Uri.parse_uri


class HttpResponse(object):
    status = None
    reason = None
    _body = None

    def __init__(self, status=None, reason=None, headers=None, body=None):
        self._headers = headers or {}
        if status is not None:
            self.status = status
        if reason is not None:
            self.reason = reason
        if body is not None:
            if hasattr(body, 'read'):
                self._body = body
            else:
                self._body = io.StringIO(body)

    def getheader(self, name, default=None):
        if name in self._headers:
            return self._headers[name]
        else:
            return default

    def getheaders(self):
        return self._headers

    def read(self, amt=None):
        if self._body is None:
            return None
        if not amt:
            return self._body.read()
        else:
            return self._body.read(amt)


def _dump_response(http_response):
    """Converts to a string for printing debug messages.

    Does not read the body since that may consume the content.
    """
    output = 'HttpResponse\n  status: %s\n  reason: %s\n  headers:' % (
        http_response.status, http_response.reason)
    headers = get_headers(http_response)
    if isinstance(headers, dict):
        for header, value in headers.items():
            output += '    %s: %s\n' % (header, value)
    else:
        for pair in headers:
            output += '    %s: %s\n' % (pair[0], pair[1])
    return output


class HttpClient(object):
    """Performs HTTP requests using httplib."""
    debug = None

    def request(self, http_request):
        return self._http_request(http_request.method, http_request.uri,
                                  http_request.headers, http_request._body_parts)

    Request = request

    def _get_connection(self, uri, headers=None):
        """Opens a socket connection to the server to set up an HTTP request.

        Args:
          uri: The full URL for the request as a Uri object.
          headers: A dict of string pairs containing the HTTP headers for the
              request.
        """
        connection = None
        if uri.scheme == 'https':
            if not uri.port:
                connection = http.client.HTTPSConnection(uri.host)
            else:
                connection = http.client.HTTPSConnection(uri.host, int(uri.port))
        else:
            if not uri.port:
                connection = http.client.HTTPConnection(uri.host)
            else:
                connection = http.client.HTTPConnection(uri.host, int(uri.port))
        return connection

    def _http_request(self, method, uri, headers=None, body_parts=None):
        """Makes an HTTP request using httplib.

        Args:
          method: str example: 'GET', 'POST', 'PUT', 'DELETE', etc.
          uri: str or atom.http_core.Uri
          headers: dict of strings mapping to strings which will be sent as HTTP
                   headers in the request.
          body_parts: list of strings, objects with a read method, or objects
                      which can be converted to strings using str. Each of these
                      will be sent in order as the body of the HTTP request.
        """
        if isinstance(uri, str):
            uri = Uri.parse_uri(uri)

        connection = self._get_connection(uri, headers=headers)

        if self.debug:
            connection.debuglevel = 1

        if connection.host != uri.host:
            connection.putrequest(method, str(uri))
        else:
            connection.putrequest(method, uri._get_relative_path())

        # Overcome a bug in Python 2.4 and 2.5
        # httplib.HTTPConnection.putrequest adding
        # HTTP request header 'Host: www.google.com:443' instead of
        # 'Host: www.google.com', and thus resulting the error message
        # 'Token invalid - AuthSub token has wrong scope' in the HTTP response.
        if (uri.scheme == 'https' and int(uri.port or 443) == 443 and
                hasattr(connection, '_buffer') and
                isinstance(connection._buffer, list)):
            header_line = 'Host: %s:443' % uri.host
            replacement_header_line = 'Host: %s' % uri.host
            try:
                connection._buffer[connection._buffer.index(header_line)] = (
                    replacement_header_line)
            except ValueError:  # header_line missing from connection._buffer
                pass

        # Send the HTTP headers.
        for header_name, value in headers.items():
            connection.putheader(header_name, value)
        connection.endheaders()

        # If there is data, send it in the request.
        if body_parts and [x for x in body_parts if x != '']:
            for part in body_parts:
                _send_data_part(part, connection)

        # Return the HTTP Response from the server.
        return connection.getresponse()


def _send_data_part(data, connection):
    if isinstance(data, str):
        # I might want to just allow str, not unicode.
        connection.send(data.encode())
    # Check to see if data is a file-like object that has a read method.
    elif hasattr(data, 'read'):
        # Read the file and send it a chunk at a time.
        while 1:
            binarydata = data.read(100000)
            if binarydata == '': break
            connection.send(binarydata)
    else:
        # The data object was not a file.
        # Try to convert to a string and send the data.
        connection.send(data)


class ProxiedHttpClient(HttpClient):
    def _get_connection(self, uri, headers=None):
        # Check to see if there are proxy settings required for this request.
        proxy = None
        if uri.scheme == 'https':
            proxy = os.environ.get('https_proxy')
        elif uri.scheme == 'http':
            proxy = os.environ.get('http_proxy')
        if not proxy:
            return HttpClient._get_connection(self, uri, headers=headers)
        # Now we have the URL of the appropriate proxy server.
        # Get a username and password for the proxy if required.
        proxy_auth = _get_proxy_auth()
        if uri.scheme == 'https':
            import socket
            if proxy_auth:
                proxy_auth = 'Proxy-Authorization: %s' % proxy_auth
            # Construct the proxy connect command.
            port = uri.port
            if not port:
                port = 443
            proxy_connect = 'CONNECT %s:%s HTTP/1.0\r\n' % (uri.host, port)
            # Set the user agent to send to the proxy
            user_agent = ''
            if headers and 'User-Agent' in headers:
                user_agent = 'User-Agent: %s\r\n' % (headers['User-Agent'])
            proxy_pieces = '%s%s%s\r\n' % (proxy_connect, proxy_auth, user_agent)
            # Find the proxy host and port.
            proxy_uri = Uri.parse_uri(proxy)
            if not proxy_uri.port:
                proxy_uri.port = '80'
            # Connect to the proxy server, very simple recv and error checking
            p_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            p_sock.connect((proxy_uri.host, int(proxy_uri.port)))
            #p_sock.sendall(proxy_pieces)
            p_sock.sendall(proxy_pieces.encode('utf-8'))
            response = ''
            # Wait for the full response.
            while response.find("\r\n\r\n") == -1:
                #response += p_sock.recv(8192)
                response += p_sock.recv(8192).decode('utf-8')
            p_status = response.split()[1]
            if p_status != str(200):
                raise ProxyError('Error status=%s' % str(p_status))
            # Trivial setup for ssl socket.
            sslobj = None
            if ssl is not None:
                context = ssl.SSLContext(ssl.PROTOCOL_TLS)
                context.minimum_version = ssl.TLSVersion.TLSv1_2
                sslobj = context.wrap_socket(p_sock, server_hostname=uri.host)
            else:
                sock_ssl = socket.ssl(p_sock, None, None)
                sslobj = http.client.FakeSocket(p_sock, sock_ssl)
            # Initalize httplib and replace with the proxy socket.
            connection = http.client.HTTPConnection(proxy_uri.host)
            connection.sock = sslobj
            return connection
        elif uri.scheme == 'http':
            proxy_uri = Uri.parse_uri(proxy)
            if not proxy_uri.port:
                proxy_uri.port = '80'
            if proxy_auth:
                headers['Proxy-Authorization'] = proxy_auth.strip()
            return http.client.HTTPConnection(proxy_uri.host, int(proxy_uri.port))
        return None


def _get_proxy_auth():
    import base64
    proxy_username = os.environ.get('proxy-username')
    if not proxy_username:
        proxy_username = os.environ.get('proxy_username')
    proxy_password = os.environ.get('proxy-password')
    if not proxy_password:
        proxy_password = os.environ.get('proxy_password')
    if proxy_username:
        user_auth = base64.b64encode(('%s:%s' % (proxy_username, proxy_password)).encode('utf-8'))
        return 'Basic %s\r\n' % (user_auth.strip().decode('utf-8'))
    else:
        return ''



# -*- coding: utf-8 -*-

# Copyright (C) 2024 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GAM action processing

"""

class GamAction():

# Keys into NAMES; arbitrary values but must be unique
  ACCEPT = 'acpt'
  ADD = 'add '
  ADD_PREVIEW = 'addp'
  APPEND = 'apnd'
  APPROVE = 'aprv'
  ARCHIVE = 'arch'
  BACKUP = 'back'
  BLOCK = 'blok'
  CANCEL = 'canc'
  CANCEL_WIPE = 'canw'
  CHECK = 'chek'
  CLAIM = 'clai'
  CLAIM_OWNERSHIP = 'clow'
  CLEAR = 'clea'
  CLOSE = 'clos'
  COLLECT = 'coll'
  COMMENT = 'comm'
  COPY = 'copy'
  COPY_MERGE = 'copm'
  CREATE = 'crea'
  CREATE_PREVIEW = 'crep'
  CREATE_SHORTCUT = 'crsc'
  DEDUP = 'dedu'
  DELETE = 'dele'
  DELETE_EMPTY = 'delm'
  DELETE_PREVIEW = 'delp'
  DELETE_SHORTCUT = 'desc'
  DEPROVISION = 'depr'
  DISABLE = 'disa'
  DOWNLOAD = 'down'
  DRAFT = 'draf'
  EMPTY = 'empt'
  ENABLE = 'enbl'
  END = 'end '
  EXISTS = 'exis'
  EXPORT = 'expo'
  EXTRACT = 'extr'
  GET_COMMAND_RESULT = 'gtcr'
  FETCH = 'fetc'
  FORWARD = 'forw'
  HIDE = 'hide'
  IMPORT = 'impo'
  INFO = 'info'
  INITIALIZE = 'init'
  INSERT = 'insr'
  INVALIDATE = 'inva'
  ISSUE_COMMAND = 'isco'
  LIST = 'list'
  LOOKUP = 'look'
  MERGE = 'merg'
  MODIFY = 'modi'
  MOVE = 'move'
  MOVE_MERGE = 'movm'
  NOACTION = 'noac'
  NOACTION_PREVIEW = 'noap'
  OBLITERATE = 'obli'
  PERFORM = 'perf'
  PRE_PROVISIONED_DISABLE ='ppdi'
  PRE_PROVISIONED_REENABLE ='ppre'
  PRINT = 'prin'
  PROCESS = 'proc'
  PROCESS_PREVIEW = 'prop'
  PURGE = 'purg'
  RECREATE = 'recr'
  REENABLE = 'reen'
  REFRESH = 'refr'
  RELABEL = 'rela'
  REMOVE = 'remo'
  REMOVE_PREVIEW = 'remp'
  RENAME = 'rena'
  REOPEN = 'reop'
  REPLACE = 'repl'
  REPLACE_DOMAIN = 'repd'
  REPORT = 'repo'
  RESET_YUBIKEY_PIV = 'rpiv'
  RESPOND = 'resp'
  RESTORE = 'rest'
  RESUBMIT = 'res'
  RETAIN = 'reta'
  RETRIEVE_DATA = 'retd'
  REVOKE = 'revo'
  SAVE = 'save'
  SEND = 'send'
  SENDEMAIL = 'snem'
  SET = 'set '
  SETUP = 'setu'
  SHARE = 'shar'
  SHOW = 'show'
  SIGNOUT = 'siou'
  SKIP = 'skip'
  SPAM = 'spam'
  SUBMIT = 'subm'
  SUSPEND = 'susp'
  SYNC = 'sync'
  TRANSFER = 'tran'
  TRANSFER_OWNERSHIP = 'trow'
  TRASH = 'tras'
  TURNOFF2SV = 'to2s'
  UNDELETE = 'unde'
  UNHIDE = 'unhi'
  UNSUSPEND = 'unsu'
  UNTRASH = 'untr'
  UPDATE = 'upda'
  UPDATE_MOVE = 'upmo'
  UPDATE_OWNER = 'updo'
  UPDATE_PREVIEW = 'updp'
  UPLOAD = 'uplo'
  UNZIP = 'unzi'
  USE = 'use '
  VERIFY = 'vrfy'
  WAITFORMAILBOX = 'wamb'
  WATCH = 'watc'
  WIPE = 'wipe'
  WIPE_PREVIEW = 'wipp'
  # Usage:
  # ACTION_NAMES[1] n Items - Delete 10 Users
  # Item xxx ACTION_NAMES[0] - User xxx Deleted
  # These values can be translated into other languages
  _NAMES = {
    ACCEPT: ['Accepted', 'Accept'],
    ADD: ['Added', 'Add'],
    ADD_PREVIEW: ['Added (Preview)', 'Add (Preview)'],
    APPEND: ['Appended', 'Append'],
    APPROVE: ['Approved', 'Approve'],
    ARCHIVE: ['Archived', 'Archive'],
    BACKUP: ['Backed up', 'Backup'],
    BLOCK: ['Blocked', 'Block'],
    CANCEL: ['Cancelled', 'Cancel'],
    CANCEL_WIPE: ['Wipe Cancelled', 'Cancel Wipe'],
    CHECK: ['Checked', 'Check'],
    CLAIM: ['Claimed', 'Claim'],
    CLAIM_OWNERSHIP: ['Ownership Claimed', 'Claim Ownership'],
    CLEAR: ['Cleared', 'Clear'],
    CLOSE: ['Closed', 'Close'],
    COLLECT: ['Collected', 'Collect'],
    COMMENT: ['Commented', 'Comment'],
    COPY: ['Copied', 'Copy'],
    COPY_MERGE: ['Copied(Merge)', 'Copy(Merge)'],
    CREATE: ['Created', 'Create'],
    CREATE_PREVIEW: ['Created (Preview)', 'Create (Preview)'],
    CREATE_SHORTCUT: ['Created Shortcut', 'Create Shortcut'],
    DEDUP: ['Duplicates Deleted', 'Delete Duplicates'],
    DELETE: ['Deleted', 'Delete'],
    DELETE_EMPTY: ['Deleted', 'Delete Empty'],
    DELETE_PREVIEW: ['Deleted (Preview)', 'Delete (Preview)'],
    DELETE_SHORTCUT: ['Deleted Shortcut', 'Delete Shortcut'],
    DEPROVISION: ['Deprovisioned', 'Deprovision'],
    DISABLE: ['Disabled', 'Disable'],
    DOWNLOAD: ['Downloaded', 'Download'],
    DRAFT: ['Drafted', 'Draft'],
    EMPTY: ['Emptied', 'Empty'],
    ENABLE: ['Enabled', 'Enable'],
    END: ['Ended', 'End'],
    EXISTS: ['Exists', 'Exists'],
    EXPORT: ['Exported', 'Export'],
    EXTRACT: ['Extracted', 'Extract'],
    FORWARD: ['Forwarded', 'Forward'],
    GET_COMMAND_RESULT: ['Got Command Result', 'Get Command Result'],
    HIDE: ['Hidden', 'Hide'],
    IMPORT: ['Imported', 'Import'],
    INFO: ['Shown', 'Show Info'],
    INITIALIZE: ['Initialized', 'Initialize'],
    INSERT: ['Inserted', 'Insert'],
    INVALIDATE: ['Invalidated', 'Invalidate'],
    ISSUE_COMMAND: ['Command Issued', 'Issue Command'],
    LIST: ['Listed', 'List'],
    LOOKUP: ['Lookedup', 'Lookup'],
    MERGE: ['Merged', 'Merge'],
    MODIFY: ['Modified', 'Modify'],
    MOVE: ['Moved', 'Move'],
    MOVE_MERGE: ['Moved(Merge)', 'Move(Merge)'],
    NOACTION: ['No Action', 'No Action'],
    NOACTION_PREVIEW: ['No Action (Preview)', 'No Action (Preview)'],
    OBLITERATE: ['Obliterated', 'Obliterate'],
    PERFORM: ['Action Performed', 'Perform Action'],
    PRE_PROVISIONED_DISABLE: ['PreProvisioned Disabled', 'PreProvisioned Disable'],
    PRE_PROVISIONED_REENABLE: ['PreProvisioned Reenabled', 'PreProvisioned Reenable'],
    PRINT: ['Printed', 'Print'],
    PROCESS: ['Processed', 'Process'],
    PROCESS_PREVIEW: ['Processed (Preview)', 'Process (Preview)'],
    PURGE: ['Purged', 'Purge'],
    RECREATE: ['Recreated', 'Recreate'],
    REENABLE: ['Reenabled', 'Reenable'],
    REFRESH: ['Refreshed', 'Refresh'],
    RELABEL: ['Relabeled', 'Relabel'],
    REMOVE: ['Removed', 'Remove'],
    REMOVE_PREVIEW: ['Removed (Preview)', 'Remove (Preview)'],
    RENAME: ['Renamed', 'Rename'],
    REOPEN: ['Reopened', 'Reopen'],
    REPLACE: ['Replaced', 'Replace'],
    REPLACE_DOMAIN: ['Domain Replaced', 'Replace Domain'],
    REPORT: ['Reported', 'Report'],
    RESET_YUBIKEY_PIV: ['Yubikey PIV Reset', 'Reset Yubikey PIV'],
    RESPOND: ['Responded', 'Respond'],
    RESTORE: ['Restored', 'Restore'],
    RESUBMIT: ['Resubmitted', 'Resubmit'],
    RETAIN: ['Retained', 'Retain'],
    RETRIEVE_DATA: ['Data Retrieved', 'Retrieve Data'],
    REVOKE: ['Revoked', 'Revoke'],
    SAVE: ['Saved', 'Save'],
    SEND: ['Sent', 'Send'],
    SENDEMAIL: ['Email Sent', 'Send Email'],
    SET: ['Set', 'Set'],
    SETUP: ['Set Up', 'Set Up'],
    SHARE: ['Shared', 'Share'],
    SHOW: ['Shown', 'Show'],
    SIGNOUT: ['Signed Out', 'Signout'],
    SKIP: ['Skipped', 'Skip'],
    SPAM: ['Marked as Spam', 'Mark as Spam'],
    SUBMIT: ['Submitted', 'Submit'],
    SUSPEND: ['Suspended', 'Suspend'],
    SYNC: ['Synced', 'Sync'],
    TRANSFER: ['Transferred', 'Transfer'],
    TRANSFER_OWNERSHIP: ['Ownership Transferred', 'Transfer Ownership'],
    TRASH: ['Trashed', 'Trash'],
    TURNOFF2SV: ['2-Step Verification Turned Off', 'Turn Off 2-Step Verification'],
    UNDELETE: ['Undeleted', 'Undelete'],
    UNHIDE: ['Unhidden', 'Unhide'],
    UNSUSPEND: ['Unsuspended', 'Unsuspend'],
    UNTRASH: ['Untrashed', 'Untrash'],
    UNZIP: ['Unzipped', 'Unzip'],
    UPDATE: ['Updated', 'Update'],
    UPDATE_MOVE: ['Updated/Moved', 'Update/Move'],
    UPDATE_OWNER: ['Updated to Owner', 'Update to Owner'],
    UPDATE_PREVIEW: ['Updated (Preview)', 'Update (Preview)'],
    UPLOAD: ['Uploaded', 'Upload'],
    USE: ['Used', 'Use'],
    VERIFY: ['Verified', 'Verify'],
    WAITFORMAILBOX: ['Mailbox is Setup', 'Check Mailbox is Setup'],
    WATCH: ['Watched', 'Watch'],
    WIPE: ['Wiped', 'Wipe'],
    WIPE_PREVIEW: ['Wiped (Preview)', 'Wipe (Preview)'],
    }
  #
  MODIFIER_CONTENTS_WITH = 'contents with'
  MODIFIER_FOR = 'for'
  MODIFIER_FROM = 'from'
  MODIFIER_IN = 'in'
  MODIFIER_INTO = 'into'
  MODIFIER_PREVIOUSLY_IN = 'previously in'
  MODIFIER_TO = 'to'
  MODIFIER_WITH_COTEACHER_OWNER = 'with co-teacher as owner'
  MODIFIER_WITH_NEW_TEACHER_OWNER = 'with new teacher as owner'
  MODIFIER_WITH_CURRENT_OWNER = 'with current owner'
  MODIFIER_WITH = 'with'
  MODIFIER_WITH_CONTENT_FROM = 'with content from'
  PREFIX_NOT = 'Not'
  PREVIEW = 'Preview'
  SUCCESS = 'Success'
  SUFFIX_FAILED = 'Failed'

  def __init__(self):
    self.action = None

  def Set(self, action):
    self.action = action

  def Get(self):
    return self.action

  def ToPerform(self):
    return self._NAMES[self.action][1]

  def Performed(self):
    return self._NAMES[self.action][0]

  def Failed(self):
    return f'{self._NAMES[self.action][1]} {self.SUFFIX_FAILED}'

  def NotPerformed(self):
    actionWords = self._NAMES[self.action][0].split(' ')
    if len(actionWords) != 2:
      return f'{self.PREFIX_NOT} {self._NAMES[self.action][0]}'
    return f'{actionWords[0]} {self.PREFIX_NOT} {actionWords[1]}'

  def PerformedName(self, action):
    return self._NAMES[action][0]

  def ToPerformName(self, action):
    return self._NAMES[action][1]

  def csvFormat(self):
    return self.action == self.PRINT



# -*- coding: utf-8 -*-

# Copyright (C) 2023 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""YubiKey"""

import base64
import datetime
from secrets import SystemRandom
import string
import sys

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from smartcard.Exceptions import CardConnectionException
from ykman.device import list_all_devices
from ykman.piv import generate_self_signed_certificate, generate_chuid
from yubikit.piv import DEFAULT_MANAGEMENT_KEY, \
                        InvalidPinError, \
                        KEY_TYPE, \
                        MANAGEMENT_KEY_TYPE, \
                        PIN_POLICY, \
                        PivSession, \
                        OBJECT_ID, \
                        SLOT, \
                        TOUCH_POLICY
from yubikit.core.smartcard import ApduError, SmartCardConnection

YUBIKEY_CONNECTION_ERROR_RC = 80
YUBIKEY_INVALID_KEY_TYPE_RC = 81
YUBIKEY_INVALID_SLOT_RC = 82
YUBIKEY_INVALID_PIN_RC = 83
YUBIKEY_APDU_ERROR_RC = 84
YUBIKEY_VALUE_ERROR_RC = 85
YUBIKEY_MULTIPLE_CONNECTED_RC = 86
YUBIKEY_NOT_FOUND_RC = 87

from gam import mplock

from gam import systemErrorExit
from gam import readStdin
from gam import writeStdout

from gam.gamlib import glmsgs as Msg

PIN_PUK_CHARS = string.ascii_letters+string.digits+string.punctuation

class YubiKey():

  def __init__(self, service_account_info=None):
    self.key_type = None
    self.slot = None
    self.serial_number = None
    self.pin = None
    self.key_id = None
    if service_account_info:
      key_type = service_account_info.get('yubikey_key_type', 'RSA2048')
      try:
        self.key_type = getattr(KEY_TYPE, key_type.upper())
      except AttributeError:
        systemErrorExit(YUBIKEY_INVALID_KEY_TYPE_RC, f'{key_type} is not a valid value for yubikey_key_type')
      slot = service_account_info.get('yubikey_slot', 'AUTHENTICATION')
      try:
        self.slot = getattr(SLOT, slot.upper())
      except AttributeError:
        systemErrorExit(YUBIKEY_INVALID_SLOT_RC, f'{slot} is not a valid value for yubikey_slot')
      self.serial_number = service_account_info.get('yubikey_serial_number')
      self.pin = service_account_info.get('yubikey_pin')
      self.key_id = service_account_info.get('private_key_id')

  def _connect(self):
    try:
      devices = list_all_devices()
      for (device, info) in devices:
        if info.serial == self.serial_number:
          return device.open_connection(SmartCardConnection)
    except CardConnectionException as err:
      systemErrorExit(YUBIKEY_CONNECTION_ERROR_RC, f'YubiKey - {err}')

  def get_certificate(self):
    try:
      conn = self._connect()
      with conn:
        session = PivSession(conn)
        if self.pin:
          try:
            session.verify_pin(self.pin)
          except InvalidPinError as err:
            systemErrorExit(YUBIKEY_INVALID_PIN_RC, f'YubiKey - {err}')
        try:
          cert = session.get_certificate(self.slot)
        except ApduError as err:
          systemErrorExit(YUBIKEY_APDU_ERROR_RC, f'YubiKey - {err}')
      cert_pem = cert.public_bytes(serialization.Encoding.PEM).decode()
      publicKeyData = base64.b64encode(cert_pem.encode())
      if isinstance(publicKeyData, bytes):
        publicKeyData = publicKeyData.decode()
      return publicKeyData
    except ValueError as err:
      systemErrorExit(YUBIKEY_VALUE_ERROR_RC, f'YubiKey - {err}')
    except TypeError as err:
      systemErrorExit(YUBIKEY_NOT_FOUND_RC, f'YubiKey - {err} - {Msg.IS_YUBIKEY_INSERTED}')

  def get_serial_number(self):
    try:
      devices = list_all_devices()
      if not devices:
        systemErrorExit(YUBIKEY_NOT_FOUND_RC, Msg.COULD_NOT_FIND_ANY_YUBIKEY)
      if self.serial_number:
        for (_, info) in devices:
          if info.serial == self.serial_number:
            return info.serial
        systemErrorExit(YUBIKEY_NOT_FOUND_RC, Msg.COULD_NOT_FIND_YUBIKEY_WITH_SERIAL.format(self.serial_number))
      if len(devices) > 1:
        serials = ', '.join([str(info.serial) for (_, info) in devices])
        systemErrorExit(YUBIKEY_MULTIPLE_CONNECTED_RC, Msg.MULTIPLE_YUBIKEYS_CONNECTED.format(serials))
      return devices[0][1].serial
    except ValueError as err:
      systemErrorExit(YUBIKEY_VALUE_ERROR_RC, f'YubiKey - {err}')

  def reset_piv(self):
    '''Resets YubiKey PIV app and generates new key for GAM to use.'''
    reply = str(readStdin(Msg.CONFIRM_WIPE_YUBIKEY_PIV).lower().strip())
    if reply != 'y':
      sys.exit(1)
    try:
      conn = self._connect()
      with conn:
        piv = PivSession(conn)
        piv.reset()
        rnd = SystemRandom()
        new_puk = ''.join(rnd.choice(PIN_PUK_CHARS) for _ in range(8))
        new_pin = ''.join(rnd.choice(PIN_PUK_CHARS) for _ in range(8))
        piv.change_puk('12345678', new_puk)
        piv.change_pin('123456', new_pin)
        writeStdout(Msg.YUBIKEY_PIN_SET_TO.format(new_pin))
        piv.authenticate(MANAGEMENT_KEY_TYPE.TDES, DEFAULT_MANAGEMENT_KEY)
        piv.verify_pin(new_pin)
        writeStdout(Msg.YUBIKEY_GENERATING_NONEXPORTABLE_PRIVATE_KEY)
        pubkey = piv.generate_key(SLOT.AUTHENTICATION,
                                  KEY_TYPE.RSA2048,
                                  PIN_POLICY.ALWAYS,
                                  TOUCH_POLICY.NEVER)
        now = datetime.datetime.utcnow()
        valid_to = now + datetime.timedelta(days=36500)
        subject = 'CN=GAM Created Key'
        piv.authenticate(MANAGEMENT_KEY_TYPE.TDES, DEFAULT_MANAGEMENT_KEY)
        piv.verify_pin(new_pin)
        cert = generate_self_signed_certificate(piv,
                                                SLOT.AUTHENTICATION,
                                                pubkey,
                                                subject,
                                                now,
                                                valid_to)
        piv.put_certificate(SLOT.AUTHENTICATION, cert)
        piv.put_object(OBJECT_ID.CHUID, generate_chuid())
    except ValueError as err:
      systemErrorExit(YUBIKEY_VALUE_ERROR_RC, f'YubiKey - {err}')
    except TypeError as err:
      systemErrorExit(YUBIKEY_NOT_FOUND_RC, f'YubiKey - {err} - {Msg.IS_YUBIKEY_INSERTED}')

  def sign(self, message):
    if mplock is not None:
      mplock.acquire()
    try:
      conn = self._connect()
      with conn:
        session = PivSession(conn)
        if self.pin:
          try:
            session.verify_pin(self.pin)
          except InvalidPinError as err:
            systemErrorExit(YUBIKEY_INVALID_PIN_RC, f'YubiKey - {err}')
        try:
          signed = session.sign(slot=self.slot,
                                key_type=self.key_type,
                                message=message,
                                hash_algorithm=hashes.SHA256(),
                                padding=padding.PKCS1v15())
        except ApduError as err:
          systemErrorExit(YUBIKEY_APDU_ERROR_RC, f'YubiKey - {err}')
    except ValueError as err:
      systemErrorExit(YUBIKEY_VALUE_ERROR_RC, f'YubiKey - {err}')
    except TypeError as err:
      systemErrorExit(YUBIKEY_NOT_FOUND_RC, f'YubiKey - {err} - {Msg.IS_YUBIKEY_INSERTED}')
    if mplock is not None:
      mplock.release()
    return signed



#!/usr/bin/env python3

from packaging import version
import sys

a = sys.argv[1]
b = sys.argv[2]
result = version.parse(a) >= version.parse(b)
if result:
  print('OK: %s is equal or newer than %s' % (a, b))
else:
  print('ERROR: %s is older than %s' % (a, b))
sys.exit(not result)



# -*- coding: utf-8 -*-

# Copyright (C) 2025 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GAM GAPI resources

"""
# callGAPI throw reasons
ABORTED = 'aborted'
ABUSIVE_CONTENT_RESTRICTION = 'abusiveContentRestriction'
ACCESS_NOT_CONFIGURED = 'accessNotConfigured'
ALREADY_EXISTS = 'alreadyExists'
APPLY_LABEL_FORBIDDEN = 'applyLabelForbidden'
AUTH_ERROR = 'authError'
BACKEND_ERROR = 'backendError'
BAD_GATEWAY = 'badGateway'
BAD_REQUEST = 'badRequest'
CANNOT_ADD_PARENT = 'cannotAddParent'
CANNOT_CHANGE_ORGANIZER = 'cannotChangeOrganizer'
CANNOT_CHANGE_ORGANIZER_OF_INSTANCE = 'cannotChangeOrganizerOfInstance'
CANNOT_CHANGE_OWN_ACL = 'cannotChangeOwnAcl'
CANNOT_CHANGE_OWNER_ACL = 'cannotChangeOwnerAcl'
CANNOT_CHANGE_OWN_PRIMARY_SUBSCRIPTION = 'cannotChangeOwnPrimarySubscription'
CANNOT_COPY_FILE = 'cannotCopyFile'
CANNOT_DELETE_ONLY_REVISION = 'cannotDeleteOnlyRevision'
CANNOT_DELETE_PERMISSION = 'cannotDeletePermission'
CANNOT_DELETE_PRIMARY_CALENDAR = 'cannotDeletePrimaryCalendar'
CANNOT_DELETE_PRIMARY_SENDAS = 'cannotDeletePrimarySendAs'
CANNOT_DELETE_RESOURCE_WITH_CHILDREN = 'cannotDeleteResourceWithChildren'
CANNOT_MODIFY_INHERITED_TEAMDRIVE_PERMISSION = 'cannotModifyInheritedTeamDrivePermission'
CANNOT_MODIFY_RESTRICTED_LABEL = 'cannotModifyRestrictedLabel'
CANNOT_MODIFY_VIEWERS_CAN_COPY_CONTENT = 'cannotModifyViewersCanCopyContent'
CANNOT_MOVE_TRASHED_ITEM_INTO_TEAMDRIVE = 'cannotMoveTrashedItemIntoTeamDrive'
CANNOT_MOVE_TRASHED_ITEM_OUT_OF_TEAMDRIVE = 'cannotMoveTrashedItemOutOfTeamDrive'
CANNOT_REMOVE_OWNER = 'cannotRemoveOwner'
CANNOT_SET_EXPIRATION = 'cannotSetExpiration'
CANNOT_SET_EXPIRATION_ON_ANYONE_OR_DOMAIN = 'cannotSetExpirationOnAnyoneOrDomain'
CANNOT_SHARE_GROUPS_WITHLINK = 'cannotShareGroupsWithLink'
CANNOT_SHARE_USERS_WITHLINK = 'cannotShareUsersWithLink'
CANNOT_SHARE_TEAMDRIVE_TOPFOLDER_WITH_ANYONEORDOMAINS = 'cannotShareTeamDriveTopFolderWithAnyoneOrDomains'
CANNOT_SHARE_TEAMDRIVE_WITH_NONGOOGLE_ACCOUNTS = 'cannotShareTeamDriveWithNonGoogleAccounts'
CANNOT_UPDATE_PERMISSION = 'cannotUpdatePermission'
CONDITION_NOT_MET = 'conditionNotMet'
CONFLICT = 'conflict'
CONTENT_OWNER_ACCOUNT_NOT_FOUND = 'contentOwnerAccountNotFound'
CROSS_DOMAIN_MOVE_RESTRICTION = 'crossDomainMoveRestriction'
CUSTOMER_EXCEEDED_ROLE_ASSIGNMENTS_LIMIT = 'CUSTOMER_EXCEEDED_ROLE_ASSIGNMENTS_LIMIT'
CUSTOMER_NOT_FOUND = 'customerNotFound'
CYCLIC_MEMBERSHIPS_NOT_ALLOWED = 'cyclicMembershipsNotAllowed'
DAILY_LIMIT_EXCEEDED = 'dailyLimitExceeded'
DELETED = 'deleted'
DELETED_USER_NOT_FOUND = 'deletedUserNotFound'
DOMAIN_ALIAS_NOT_FOUND = 'domainAliasNotFound'
DOMAIN_CANNOT_USE_APIS = 'domainCannotUseApis'
DOMAIN_NOT_FOUND = 'domainNotFound'
DOMAIN_NOT_VERIFIED_SECONDARY = 'domainNotVerifiedSecondary'
DOMAIN_POLICY = 'domainPolicy'
DOWNLOAD_QUOTA_EXCEEDED = 'downloadQuotaExceeded'
DUPLICATE = 'duplicate'
EVENT_DURATION_EXCEEDS_LIMIT = 'eventDurationExceedsLimit'
EVENT_TYPE_RESTRICTION = 'eventTypeRestriction'
EXPIRATION_DATES_MUST_BE_IN_THE_FUTURE = 'expirationDatesMustBeInTheFuture'
EXPIRATION_DATE_NOT_ALLOWED_FOR_SHARED_DRIVE_MEMBERS = 'expirationDateNotAllowedForSharedDriveMembers'
FAILED_PRECONDITION = 'failedPrecondition'
FIELD_IN_USE = 'fieldInUse'
FIELD_NOT_WRITABLE = 'fieldNotWritable'
FILE_NEVER_WRITABLE = 'fileNeverWritable'
FILE_NOT_FOUND = 'fileNotFound'
FILE_ORGANIZER_NOT_YET_ENABLED_FOR_THIS_TEAMDRIVE = 'fileOrganizerNotYetEnabledForThisTeamDrive'
FILE_ORGANIZER_ON_FOLDERS_IN_SHARED_DRIVE_ONLY = 'fileOrganizerOnFoldersInSharedDriveOnly'
FILE_ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED = 'fileOrganizerOnNonTeamDriveNotSupported'
FILE_OWNER_NOT_MEMBER_OF_TEAMDRIVE = 'fileOwnerNotMemberOfTeamDrive'
FILE_OWNER_NOT_MEMBER_OF_WRITER_DOMAIN = 'fileOwnerNotMemberOfWriterDomain'
FILE_WRITER_TEAMDRIVE_MOVE_IN_DISABLED = 'fileWriterTeamDriveMoveInDisabled'
FORBIDDEN = 'forbidden'
GATEWAY_TIMEOUT = 'gatewayTimeout'
GROUP_NOT_FOUND = 'groupNotFound'
ILLEGAL_ACCESS_ROLE_FOR_DEFAULT = 'illegalAccessRoleForDefault'
INSUFFICIENT_ADMINISTRATOR_PRIVILEGES = 'insufficientAdministratorPrivileges'
INSUFFICIENT_ARCHIVED_USER_LICENSES = 'insufficientArchivedUserLicenses'
INSUFFICIENT_FILE_PERMISSIONS = 'insufficientFilePermissions'
INSUFFICIENT_PARENT_PERMISSIONS = 'insufficientParentPermissions'
INSUFFICIENT_PERMISSIONS = 'insufficientPermissions'
INTERNAL_ERROR = 'internalError'
INVALID = 'invalid'
INVALID_ARGUMENT = 'invalidArgument'
INVALID_ATTRIBUTE_VALUE = 'invalidAttributeValue'
INVALID_CUSTOMER_ID = 'invalidCustomerId'
INVALID_INPUT = 'invalidInput'
INVALID_LINK_VISIBILITY = 'invalidLinkVisibility'
INVALID_MEMBER = 'invalidMember'
INVALID_MESSAGE_ID = 'invalidMessageId'
INVALID_ORGUNIT = 'invalidOrgunit'
INVALID_ORGUNIT_NAME = 'invalidOrgunitName'
INVALID_OWNERSHIP_TRANSFER = 'invalidOwnershipTransfer'
INVALID_PARAMETER = 'invalidParameter'
INVALID_PARENT_ORGUNIT = 'invalidParentOrgunit'
INVALID_QUERY = 'invalidQuery'
INVALID_RESOURCE = 'invalidResource'
INVALID_SCHEMA_VALUE = 'invalidSchemaValue'
INVALID_SCOPE_VALUE = 'invalidScopeValue'
INVALID_SHARING_REQUEST = 'invalidSharingRequest'
LABEL_MULTIPLE_VALUES_FOR_SINGULAR_FIELD = 'labelMultipleValuesForSingularField'
LABEL_MUTATION_FORBIDDEN = 'labelMutationForbidden'
LABEL_MUTATION_ILLEGAL_SELECTION = 'labelMutationIllegalSelection'
LABEL_MUTATION_UNKNOWN_FIELD = 'labelMutationUnknownField'
LIMIT_EXCEEDED = 'limitExceeded'
LOGIN_REQUIRED = 'loginRequired'
MALFORMED_WORKING_LOCATION_EVENT = 'malformedWorkingLocationEvent'
MEMBER_NOT_FOUND = 'memberNotFound'
MYDRIVE_HIERARCHY_DEPTH_LIMIT_EXCEEDED = 'myDriveHierarchyDepthLimitExceeded'
NO_LIST_TEAMDRIVES_ADMINISTRATOR_PRIVILEGE = 'noListTeamDrivesAdministratorPrivilege'
NO_MANAGE_TEAMDRIVE_ADMINISTRATOR_PRIVILEGE = 'noManageTeamDriveAdministratorPrivilege'
NOT_A_CALENDAR_USER = 'notACalendarUser'
NOT_FOUND = 'notFound'
NOT_IMPLEMENTED = 'notImplemented'
NUM_CHILDREN_IN_NON_ROOT_LIMIT_EXCEEDED = 'numChildrenInNonRootLimitExceeded'
OPERATION_NOT_SUPPORTED = 'operationNotSupported'
ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED = 'organizerOnNonTeamDriveNotSupported'
ORGANIZER_ON_NON_TEAMDRIVE_ITEM_NOT_SUPPORTED = 'organizerOnNonTeamDriveItemNotSupported'
ORGUNIT_NOT_FOUND = 'orgunitNotFound'
OWNER_ON_TEAMDRIVE_ITEM_NOT_SUPPORTED = 'ownerOnTeamDriveItemNotSupported'
OWNERSHIP_CHANGE_ACROSS_DOMAIN_NOT_PERMITTED = 'ownershipChangeAcrossDomainNotPermitted'
PARTICIPANT_IS_NEITHER_ORGANIZER_NOR_ATTENDEE = 'participantIsNeitherOrganizerNorAttendee'
PERMISSION_DENIED = 'permissionDenied'
PERMISSION_NOT_FOUND = 'permissionNotFound'
PHOTO_NOT_FOUND = 'photoNotFound'
PUBLISH_OUT_NOT_PERMITTED = 'publishOutNotPermitted'
QUERY_REQUIRES_ADMIN_CREDENTIALS = 'queryRequiresAdminCredentials'
QUOTA_EXCEEDED = 'quotaExceeded'
RATE_LIMIT_EXCEEDED = 'rateLimitExceeded'
REQUIRED = 'required'
REQUIRED_ACCESS_LEVEL = 'requiredAccessLevel'
RESOURCE_EXHAUSTED = 'resourceExhausted'
RESOURCE_ID_NOT_FOUND = 'resourceIdNotFound'
RESOURCE_NOT_FOUND = 'resourceNotFound'
RESPONSE_PREPARATION_FAILURE = 'responsePreparationFailure'
REVISION_DELETION_NOT_SUPPORTED = 'revisionDeletionNotSupported'
REVISION_NOT_FOUND = 'revisionNotFound'
REVISIONS_NOT_SUPPORTED = 'revisionsNotSupported'
SERVICE_LIMIT = 'serviceLimit'
SERVICE_NOT_AVAILABLE = 'serviceNotAvailable'
SHARE_IN_NOT_PERMITTED = 'shareInNotPermitted'
SHARE_OUT_NOT_PERMITTED = 'shareOutNotPermitted'
SHARE_OUT_NOT_PERMITTED_TO_USER = 'shareOutNotPermittedToUser'
SHARE_OUT_WARNING = 'shareOutWarning'
SHARING_RATE_LIMIT_EXCEEDED = 'sharingRateLimitExceeded'
SHORTCUT_TARGET_INVALID = 'shortcutTargetInvalid'
STORAGE_QUOTA_EXCEEDED = 'storageQuotaExceeded'
SYSTEM_ERROR = 'systemError'
TARGET_USER_ROLE_LIMITED_BY_LICENSE_RESTRICTION = 'targetUserRoleLimitedByLicenseRestriction'
TEAMDRIVE_ALREADY_EXISTS = 'teamDriveAlreadyExists'
TEAMDRIVE_DOMAIN_USERS_ONLY_RESTRICTION = 'teamDriveDomainUsersOnlyRestriction'
TEAMDRIVE_TEAM_MEMBERS_ONLY_RESTRICTION = 'teamDriveTeamMembersOnlyRestriction'
TEAMDRIVE_FILE_LIMIT_EXCEEDED = 'teamDriveFileLimitExceeded'
TEAMDRIVE_HIERARCHY_TOO_DEEP = 'teamDriveHierarchyTooDeep'
TEAMDRIVE_MEMBERSHIP_REQUIRED = 'teamDriveMembershipRequired'
TEAMDRIVES_FOLDER_MOVE_IN_NOT_SUPPORTED = 'teamDrivesFolderMoveInNotSupported'
TEAMDRIVES_FOLDER_SHARING_NOT_SUPPORTED = 'teamDrivesFolderSharingNotSupported'
TEAMDRIVES_PARENT_LIMIT = 'teamDrivesParentLimit'
TEAMDRIVES_SHARING_RESTRICTION_NOT_ALLOWED = 'teamDrivesSharingRestrictionNotAllowed'
TEAMDRIVES_SHORTCUT_FILE_NOT_SUPPORTED = 'teamDrivesShortcutFileNotSupported'
TIME_RANGE_EMPTY = 'timeRangeEmpty'
TRANSIENT_ERROR = 'transientError'
UNKNOWN_ERROR = 'unknownError'
UNSUPPORTED_LANGUAGE_CODE = 'unsupportedLanguageCode'
UNSUPPORTED_SUPERVISED_ACCOUNT = 'unsupportedSupervisedAccount'
UPLOAD_TOO_LARGE = 'uploadTooLarge'
USER_CANNOT_CREATE_TEAMDRIVES = 'userCannotCreateTeamDrives'
USER_ACCESS = 'userAccess'
USER_NOT_FOUND = 'userNotFound'
USER_RATE_LIMIT_EXCEEDED = 'userRateLimitExceeded'
#
DEFAULT_RETRY_REASONS = [QUOTA_EXCEEDED, RATE_LIMIT_EXCEEDED, SHARING_RATE_LIMIT_EXCEEDED, USER_RATE_LIMIT_EXCEEDED,
                         BACKEND_ERROR, BAD_GATEWAY, GATEWAY_TIMEOUT, INTERNAL_ERROR, TRANSIENT_ERROR]
SERVICE_NOT_AVAILABLE_RETRY_REASONS = [SERVICE_NOT_AVAILABLE]
ACTIVITY_THROW_REASONS = [SERVICE_NOT_AVAILABLE, BAD_REQUEST]
ALERT_THROW_REASONS = [SERVICE_NOT_AVAILABLE, AUTH_ERROR, PERMISSION_DENIED]
CALENDAR_THROW_REASONS = [SERVICE_NOT_AVAILABLE, AUTH_ERROR, NOT_A_CALENDAR_USER]
CIGROUP_CREATE_THROW_REASONS = [SERVICE_NOT_AVAILABLE, ALREADY_EXISTS, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, INVALID, INVALID_ARGUMENT, PERMISSION_DENIED, FAILED_PRECONDITION]
CIGROUP_GET_THROW_REASONS = [SERVICE_NOT_AVAILABLE, NOT_FOUND, GROUP_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, BAD_REQUEST, INVALID, SYSTEM_ERROR, PERMISSION_DENIED]
CIGROUP_LIST_THROW_REASONS = [SERVICE_NOT_AVAILABLE, RESOURCE_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, BAD_REQUEST, INVALID, INVALID_ARGUMENT, SYSTEM_ERROR, PERMISSION_DENIED]
CIGROUP_LIST_USERKEY_THROW_REASONS = CIGROUP_LIST_THROW_REASONS+[INVALID_ARGUMENT]
CIGROUP_UPDATE_THROW_REASONS = [SERVICE_NOT_AVAILABLE, NOT_FOUND, GROUP_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS,
                                FORBIDDEN, BAD_REQUEST, INVALID, INVALID_INPUT, INVALID_ARGUMENT,
                                SYSTEM_ERROR, PERMISSION_DENIED, FAILED_PRECONDITION]
CIGROUP_RETRY_REASONS = [INVALID, SYSTEM_ERROR, SERVICE_NOT_AVAILABLE]
CIMEMBERS_THROW_REASONS = [SERVICE_NOT_AVAILABLE, MEMBER_NOT_FOUND, INVALID_MEMBER]
CISSO_CREATE_THROW_REASONS = [SERVICE_NOT_AVAILABLE, FAILED_PRECONDITION, NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, INVALID, INVALID_ARGUMENT, PERMISSION_DENIED, INTERNAL_ERROR]
CISSO_GET_THROW_REASONS = [SERVICE_NOT_AVAILABLE, NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, BAD_REQUEST, INVALID, SYSTEM_ERROR, PERMISSION_DENIED, INTERNAL_ERROR]
CISSO_LIST_THROW_REASONS = [SERVICE_NOT_AVAILABLE, NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, BAD_REQUEST, INVALID, SYSTEM_ERROR, PERMISSION_DENIED, INTERNAL_ERROR]
CISSO_UPDATE_THROW_REASONS = [SERVICE_NOT_AVAILABLE, NOT_FOUND, FAILED_PRECONDITION, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS,
                              FORBIDDEN, BAD_REQUEST, INVALID, INVALID_INPUT, INVALID_ARGUMENT,
                              SYSTEM_ERROR, PERMISSION_DENIED, INTERNAL_ERROR]
CONTACT_DELEGATE_THROW_REASONS = [SERVICE_NOT_AVAILABLE, BAD_REQUEST, FAILED_PRECONDITION, PERMISSION_DENIED, FORBIDDEN, INVALID_ARGUMENT]
COURSE_ACCESS_THROW_REASONS = [NOT_FOUND, INSUFFICIENT_PERMISSIONS, PERMISSION_DENIED, FORBIDDEN, INVALID_ARGUMENT]
DRIVE_USER_THROW_REASONS = [SERVICE_NOT_AVAILABLE, AUTH_ERROR, DOMAIN_POLICY]
DRIVE_ACCESS_THROW_REASONS = DRIVE_USER_THROW_REASONS+[FILE_NOT_FOUND, FORBIDDEN, INTERNAL_ERROR, INSUFFICIENT_FILE_PERMISSIONS, UNKNOWN_ERROR, INVALID]
DRIVE_COPY_THROW_REASONS = DRIVE_ACCESS_THROW_REASONS+[CANNOT_COPY_FILE, BAD_REQUEST, RESPONSE_PREPARATION_FAILURE, TEAMDRIVES_SHARING_RESTRICTION_NOT_ALLOWED,
                                                       FIELD_NOT_WRITABLE, RATE_LIMIT_EXCEEDED, USER_RATE_LIMIT_EXCEEDED,
                                                       STORAGE_QUOTA_EXCEEDED, TEAMDRIVE_FILE_LIMIT_EXCEEDED, TEAMDRIVE_HIERARCHY_TOO_DEEP]
DRIVE_GET_THROW_REASONS = DRIVE_USER_THROW_REASONS+[FILE_NOT_FOUND, DOWNLOAD_QUOTA_EXCEEDED]
DRIVE3_CREATE_ACL_THROW_REASONS = [BAD_REQUEST, INVALID, INVALID_SHARING_REQUEST, OWNERSHIP_CHANGE_ACROSS_DOMAIN_NOT_PERMITTED,
                                   CANNOT_SET_EXPIRATION, CANNOT_SET_EXPIRATION_ON_ANYONE_OR_DOMAIN,
                                   EXPIRATION_DATES_MUST_BE_IN_THE_FUTURE, EXPIRATION_DATE_NOT_ALLOWED_FOR_SHARED_DRIVE_MEMBERS,
                                   NOT_FOUND, TEAMDRIVE_DOMAIN_USERS_ONLY_RESTRICTION, TEAMDRIVE_TEAM_MEMBERS_ONLY_RESTRICTION,
                                   TARGET_USER_ROLE_LIMITED_BY_LICENSE_RESTRICTION, INSUFFICIENT_ADMINISTRATOR_PRIVILEGES, SHARING_RATE_LIMIT_EXCEEDED,
                                   PUBLISH_OUT_NOT_PERMITTED, SHARE_IN_NOT_PERMITTED, SHARE_OUT_NOT_PERMITTED, SHARE_OUT_NOT_PERMITTED_TO_USER,
                                   CANNOT_SHARE_TEAMDRIVE_TOPFOLDER_WITH_ANYONEORDOMAINS,
                                   CANNOT_SHARE_TEAMDRIVE_WITH_NONGOOGLE_ACCOUNTS,
                                   OWNER_ON_TEAMDRIVE_ITEM_NOT_SUPPORTED,
                                   ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED,
                                   ORGANIZER_ON_NON_TEAMDRIVE_ITEM_NOT_SUPPORTED,
                                   FILE_ORGANIZER_NOT_YET_ENABLED_FOR_THIS_TEAMDRIVE,
                                   FILE_ORGANIZER_ON_FOLDERS_IN_SHARED_DRIVE_ONLY,
                                   FILE_ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED,
                                   TEAMDRIVES_FOLDER_SHARING_NOT_SUPPORTED, INVALID_LINK_VISIBILITY, ABUSIVE_CONTENT_RESTRICTION]
DRIVE3_GET_ACL_REASONS = DRIVE_USER_THROW_REASONS+[FILE_NOT_FOUND, FORBIDDEN, INTERNAL_ERROR,
                                                   INSUFFICIENT_ADMINISTRATOR_PRIVILEGES, INSUFFICIENT_FILE_PERMISSIONS,
                                                   UNKNOWN_ERROR, INVALID]
DRIVE3_UPDATE_ACL_THROW_REASONS = [BAD_REQUEST, INVALID_OWNERSHIP_TRANSFER, CANNOT_REMOVE_OWNER,
                                   CANNOT_SET_EXPIRATION, CANNOT_SET_EXPIRATION_ON_ANYONE_OR_DOMAIN,
                                   EXPIRATION_DATES_MUST_BE_IN_THE_FUTURE, EXPIRATION_DATE_NOT_ALLOWED_FOR_SHARED_DRIVE_MEMBERS,
                                   OWNERSHIP_CHANGE_ACROSS_DOMAIN_NOT_PERMITTED,
                                   NOT_FOUND, TEAMDRIVE_DOMAIN_USERS_ONLY_RESTRICTION, TEAMDRIVE_TEAM_MEMBERS_ONLY_RESTRICTION,
                                   TARGET_USER_ROLE_LIMITED_BY_LICENSE_RESTRICTION, INSUFFICIENT_ADMINISTRATOR_PRIVILEGES, SHARING_RATE_LIMIT_EXCEEDED,
                                   PUBLISH_OUT_NOT_PERMITTED, SHARE_IN_NOT_PERMITTED, SHARE_OUT_NOT_PERMITTED, SHARE_OUT_NOT_PERMITTED_TO_USER,
                                   CANNOT_SHARE_TEAMDRIVE_TOPFOLDER_WITH_ANYONEORDOMAINS,
                                   CANNOT_SHARE_TEAMDRIVE_WITH_NONGOOGLE_ACCOUNTS,
                                   OWNER_ON_TEAMDRIVE_ITEM_NOT_SUPPORTED,
                                   ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED,
                                   ORGANIZER_ON_NON_TEAMDRIVE_ITEM_NOT_SUPPORTED,
                                   FILE_ORGANIZER_NOT_YET_ENABLED_FOR_THIS_TEAMDRIVE,
                                   FILE_ORGANIZER_ON_FOLDERS_IN_SHARED_DRIVE_ONLY,
                                   FILE_ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED,
                                   CANNOT_UPDATE_PERMISSION,
                                   CANNOT_MODIFY_INHERITED_TEAMDRIVE_PERMISSION,
                                   FIELD_NOT_WRITABLE, PERMISSION_NOT_FOUND]
DRIVE3_DELETE_ACL_THROW_REASONS = [BAD_REQUEST, CANNOT_REMOVE_OWNER,
                                   CANNOT_MODIFY_INHERITED_TEAMDRIVE_PERMISSION,
                                   INSUFFICIENT_ADMINISTRATOR_PRIVILEGES, SHARING_RATE_LIMIT_EXCEEDED,
                                   NOT_FOUND, PERMISSION_NOT_FOUND, CANNOT_DELETE_PERMISSION]
DRIVE3_MODIFY_LABEL_THROW_REASONS = DRIVE_USER_THROW_REASONS+[FILE_NOT_FOUND, NOT_FOUND, FORBIDDEN, INTERNAL_ERROR,
                                                              FILE_NEVER_WRITABLE, APPLY_LABEL_FORBIDDEN,
                                                              INSUFFICIENT_ADMINISTRATOR_PRIVILEGES, INSUFFICIENT_FILE_PERMISSIONS,
                                                              UNKNOWN_ERROR, INVALID_INPUT, BAD_REQUEST,
                                                              LABEL_MULTIPLE_VALUES_FOR_SINGULAR_FIELD, LABEL_MUTATION_FORBIDDEN,
                                                              LABEL_MUTATION_ILLEGAL_SELECTION, LABEL_MUTATION_UNKNOWN_FIELD]
DOCS_ACCESS_THROW_REASONS = DRIVE_USER_THROW_REASONS+[NOT_FOUND, PERMISSION_DENIED, FORBIDDEN, INTERNAL_ERROR, INSUFFICIENT_FILE_PERMISSIONS,
                                                      BAD_REQUEST, INVALID, INVALID_ARGUMENT, FAILED_PRECONDITION]
GMAIL_THROW_REASONS = [SERVICE_NOT_AVAILABLE, BAD_REQUEST]
GMAIL_LIST_THROW_REASONS = [FAILED_PRECONDITION, PERMISSION_DENIED, INVALID, INVALID_ARGUMENT]
GMAIL_SMIME_THROW_REASONS = [SERVICE_NOT_AVAILABLE, BAD_REQUEST, INVALID_ARGUMENT, FORBIDDEN, NOT_FOUND, PERMISSION_DENIED]
GROUP_GET_THROW_REASONS = [GROUP_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, BAD_REQUEST, INVALID, SYSTEM_ERROR]
GROUP_GET_RETRY_REASONS = [INVALID, SYSTEM_ERROR, SERVICE_NOT_AVAILABLE]
GROUP_CREATE_THROW_REASONS = [DUPLICATE, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, INVALID, INVALID_INPUT]
GROUP_UPDATE_THROW_REASONS = [GROUP_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, INVALID, INVALID_INPUT]
GROUP_SETTINGS_THROW_REASONS = [NOT_FOUND, GROUP_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, SYSTEM_ERROR, PERMISSION_DENIED,
                                INVALID, INVALID_ARGUMENT, INVALID_PARAMETER, INVALID_ATTRIBUTE_VALUE, INVALID_INPUT,
                                SERVICE_LIMIT, SERVICE_NOT_AVAILABLE, AUTH_ERROR, REQUIRED]
GROUP_SETTINGS_RETRY_REASONS = [INVALID, SERVICE_LIMIT, SERVICE_NOT_AVAILABLE]
GROUP_LIST_THROW_REASONS = [RESOURCE_NOT_FOUND, DOMAIN_NOT_FOUND, FORBIDDEN, BAD_REQUEST]
GROUP_LIST_USERKEY_THROW_REASONS = GROUP_LIST_THROW_REASONS+[INVALID_MEMBER, INVALID_INPUT]
KEEP_THROW_REASONS = [AUTH_ERROR, BAD_REQUEST, PERMISSION_DENIED, INVALID_ARGUMENT, NOT_FOUND]
LOOKERSTUDIO_THROW_REASONS = [INVALID_ARGUMENT, SERVICE_NOT_AVAILABLE, BAD_REQUEST, NOT_FOUND, PERMISSION_DENIED, INTERNAL_ERROR]
MEMBERS_THROW_REASONS = [GROUP_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, INVALID, FORBIDDEN, SERVICE_NOT_AVAILABLE]
MEMBERS_RETRY_REASONS = [SYSTEM_ERROR, SERVICE_NOT_AVAILABLE]
ORGUNIT_GET_THROW_REASONS = [INVALID_ORGUNIT, ORGUNIT_NOT_FOUND, BACKEND_ERROR, BAD_REQUEST, INVALID_CUSTOMER_ID, LOGIN_REQUIRED]
PEOPLE_ACCESS_THROW_REASONS = [SERVICE_NOT_AVAILABLE, FORBIDDEN, PERMISSION_DENIED, FAILED_PRECONDITION]
RESELLER_THROW_REASONS = [BAD_REQUEST, RESOURCE_NOT_FOUND, FORBIDDEN, INVALID]
SHEETS_ACCESS_THROW_REASONS = DRIVE_USER_THROW_REASONS+[NOT_FOUND, PERMISSION_DENIED, FORBIDDEN, INTERNAL_ERROR, INSUFFICIENT_FILE_PERMISSIONS,
                                                        BAD_REQUEST, INVALID, INVALID_ARGUMENT, FAILED_PRECONDITION]
TASK_THROW_REASONS = [BAD_REQUEST, PERMISSION_DENIED, INVALID, NOT_FOUND, ACCESS_NOT_CONFIGURED]
TASKLIST_THROW_REASONS = [BAD_REQUEST, PERMISSION_DENIED, INVALID, NOT_FOUND, ACCESS_NOT_CONFIGURED]
USER_GET_THROW_REASONS = [USER_NOT_FOUND, DOMAIN_NOT_FOUND, DOMAIN_CANNOT_USE_APIS, FORBIDDEN, BAD_REQUEST, SYSTEM_ERROR]
YOUTUBE_THROW_REASONS = [SERVICE_NOT_AVAILABLE, AUTH_ERROR, UNSUPPORTED_SUPERVISED_ACCOUNT, UNSUPPORTED_LANGUAGE_CODE, CONTENT_OWNER_ACCOUNT_NOT_FOUND]

REASON_MESSAGE_MAP = {
  ABORTED: [
    ('Label name exists or conflicts', DUPLICATE),
    ('The operation was aborted', ABORTED),
    ],
  CONDITION_NOT_MET: [
    ('Cyclic memberships not allowed', CYCLIC_MEMBERSHIPS_NOT_ALLOWED),
    ('undelete', DELETED_USER_NOT_FOUND),
    ],
  FAILED_PRECONDITION: [
    ('Bad Request', BAD_REQUEST),
    ('Mail service not enabled', SERVICE_NOT_AVAILABLE),
    ],
  INVALID: [
    ('userId', USER_NOT_FOUND),
    ('memberKey', INVALID_MEMBER),
    ('A system error has occurred', SYSTEM_ERROR),
    ('Expiration dates must be in the future', EXPIRATION_DATES_MUST_BE_IN_THE_FUTURE),
    ('Invalid attribute value', INVALID_ATTRIBUTE_VALUE),
    ('Invalid Customer Id', INVALID_CUSTOMER_ID),
    ('Invalid Input: INVALID_OU_ID', INVALID_ORGUNIT),
    ('Invalid Input: custom_schema', INVALID_SCHEMA_VALUE),
    ('Invalid Input: groupKey', INVALID_INPUT),
    ('Invalid Input: resource', INVALID_RESOURCE),
    ('Invalid Input:', INVALID_INPUT),
    ('Invalid Input', INVALID_INPUT),
    ('Invalid Org Unit', INVALID_ORGUNIT),
    ('Invalid Ou Id', INVALID_ORGUNIT),
    ('Invalid Ou Name', INVALID_ORGUNIT_NAME),
    ('Invalid Parent Orgunit Id', INVALID_PARENT_ORGUNIT),
    ('Invalid query', INVALID_QUERY),
    ('Invalid scope value', INVALID_SCOPE_VALUE),
    ('Invalid value', INVALID_INPUT),
    ('New domain name is not a verified secondary domain', DOMAIN_NOT_VERIFIED_SECONDARY),
    ('PermissionDenied', PERMISSION_DENIED),
    ],
  INVALID_ARGUMENT: [
    ('Cannot delete primary send-as', CANNOT_DELETE_PRIMARY_SENDAS),
    ('Invalid id value', INVALID_MESSAGE_ID),
    ('Invalid ids value', INVALID_MESSAGE_ID),
    ],
  NOT_FOUND: [
    ('userKey', USER_NOT_FOUND),
    ('groupKey', GROUP_NOT_FOUND),
    ('memberKey', MEMBER_NOT_FOUND),
    ('photo', PHOTO_NOT_FOUND),
    ('resource_id', RESOURCE_ID_NOT_FOUND),
    ('resourceId', RESOURCE_ID_NOT_FOUND),
    ('Customer doesn\'t exist', CUSTOMER_NOT_FOUND),
    ('Domain alias does not exist', DOMAIN_ALIAS_NOT_FOUND),
    ('Domain not found', DOMAIN_NOT_FOUND),
    ('domain', DOMAIN_NOT_FOUND),
    ('File not found', FILE_NOT_FOUND),
    ('Org unit not found', ORGUNIT_NOT_FOUND),
    ('Permission not found', PERMISSION_NOT_FOUND),
    ('Resource Not Found', RESOURCE_NOT_FOUND),
    ('Revision not found', REVISION_NOT_FOUND),
    ('Shared Drive not found', NOT_FOUND),
    ('Not Found', NOT_FOUND),
    ],
  REQUIRED: [
    ('Login Required', LOGIN_REQUIRED),
    ('memberKey', MEMBER_NOT_FOUND),
    ],
  RESOURCE_NOT_FOUND: [
    ('resourceId', RESOURCE_ID_NOT_FOUND),
    ],
  }

class aborted(Exception):
  pass
class abusiveContentRestriction(Exception):
  pass
class accessNotConfigured(Exception):
  pass
class alreadyExists(Exception):
  pass
class applyLabelForbidden(Exception):
  pass
class authError(Exception):
  pass
class backendError(Exception):
  pass
class badRequest(Exception):
  pass
class cannotAddParent(Exception):
  pass
class cannotChangeOrganizer(Exception):
  pass
class cannotChangeOrganizerOfInstance(Exception):
  pass
class cannotChangeOwnAcl(Exception):
  pass
class cannotChangeOwnerAcl(Exception):
  pass
class cannotChangeOwnPrimarySubscription(Exception):
  pass
class cannotCopyFile(Exception):
  pass
class cannotDeleteOnlyRevision(Exception):
  pass
class cannotDeletePermission(Exception):
  pass
class cannotDeletePrimaryCalendar(Exception):
  pass
class cannotDeletePrimarySendAs(Exception):
  pass
class cannotDeleteResourceWithChildren(Exception):
  pass
class cannotModifyInheritedTeamDrivePermission(Exception):
  pass
class cannotModifyRestrictedLabel(Exception):
  pass
class cannotModifyViewersCanCopyContent(Exception):
  pass
class cannotMoveTrashedItemIntoTeamDrive(Exception):
  pass
class cannotMoveTrashedItemOutOfTeamDrive(Exception):
  pass
class cannotRemoveOwner(Exception):
  pass
class cannotSetExpiration(Exception):
  pass
class cannotSetExpirationOnAnyoneOrDomain(Exception):
  pass
class cannotShareGroupsWithLink(Exception):
  pass
class cannotShareUsersWithLink(Exception):
  pass
class cannotShareTeamDriveTopFolderWithAnyoneOrDomains(Exception):
  pass
class cannotShareTeamDriveWithNonGoogleAccounts(Exception):
  pass
class cannotUpdatePermission(Exception):
  pass
class conditionNotMet(Exception):
  pass
class conflict(Exception):
  pass
class contentOwnerAccountNotFound(Exception):
  pass
class crossDomainMoveRestriction(Exception):
  pass
class customerExceededRoleAssignmentsLimit(Exception):
  pass
class customerNotFound(Exception):
  pass
class cyclicMembershipsNotAllowed(Exception):
  pass
class deleted(Exception):
  pass
class deletedUserNotFound(Exception):
  pass
class domainAliasNotFound(Exception):
  pass
class domainCannotUseApis(Exception):
  pass
class domainNotFound(Exception):
  pass
class domainNotVerifiedSecondary(Exception):
  pass
class domainPolicy(Exception):
  pass
class downloadQuotaExceeded(Exception):
  pass
class duplicate(Exception):
  pass
class eventDurationExceedsLimit(Exception):
  pass
class eventTypeRestriction(Exception):
  pass
class expirationDatesMustBeInTheFuture(Exception):
  pass
class expirationDateNotAllowedForSharedDriveMembers(Exception):
  pass
class failedPrecondition(Exception):
  pass
class fieldInUse(Exception):
  pass
class fieldNotWritable(Exception):
  pass
class fileNeverWritable(Exception):
  pass
class fileNotFound(Exception):
  pass
class fileOrganizerNotYetEnabledForThisTeamDrive(Exception):
  pass
class fileOrganizerOnFoldersInSharedDriveOnly(Exception):
  pass
class fileOrganizerOnNonTeamDriveNotSupported(Exception):
  pass
class fileOwnerNotMemberOfTeamDrive(Exception):
  pass
class fileOwnerNotMemberOfWriterDomain(Exception):
  pass
class fileWriterTeamDriveMoveInDisabled(Exception):
  pass
class forbidden(Exception):
  pass
class groupNotFound(Exception):
  pass
class illegalAccessRoleForDefault(Exception):
  pass
class insufficientAdministratorPrivileges(Exception):
  pass
class insufficientArchivedUserLicenses(Exception):
  pass
class insufficientFilePermissions(Exception):
  pass
class insufficientParentPermissions(Exception):
  pass
class insufficientPermissions(Exception):
  pass
class internalError(Exception):
  pass
class invalid(Exception):
  pass
class invalidArgument(Exception):
  pass
class invalidAttributeValue(Exception):
  pass
class invalidCustomerId(Exception):
  pass
class invalidInput(Exception):
  pass
class invalidLinkVisibility(Exception):
  pass
class invalidMember(Exception):
  pass
class invalidMessageId(Exception):
  pass
class invalidOrgunit(Exception):
  pass
class invalidOrgunitName(Exception):
  pass
class invalidOwnershipTransfer(Exception):
  pass
class invalidParameter(Exception):
  pass
class invalidParentOrgunit(Exception):
  pass
class invalidQuery(Exception):
  pass
class invalidResource(Exception):
  pass
class invalidSchemaValue(Exception):
  pass
class invalidScopeValue(Exception):
  pass
class invalidSharingRequest(Exception):
  pass
class labelMultipleValuesForSingularField(Exception):
  pass
class labelMutationForbidden(Exception):
  pass
class labelMutationIllegalSelection(Exception):
  pass
class labelMutationUnknownField(Exception):
  pass
class limitExceeded(Exception):
  pass
class loginRequired(Exception):
  pass
class malformedWorkingLocationEvent(Exception):
  pass
class memberNotFound(Exception):
  pass
class noListTeamDrivesAdministratorPrivilege(Exception):
  pass
class noManageTeamDriveAdministratorPrivilege(Exception):
  pass
class notACalendarUser(Exception):
  pass
class notFound(Exception):
  pass
class notImplemented(Exception):
  pass
class operationNotSupported(Exception):
  pass
class organizerOnNonTeamDriveNotSupported(Exception):
  pass
class organizerOnNonTeamDriveItemNotSupported(Exception):
  pass
class orgunitNotFound(Exception):
  pass
class ownerOnTeamDriveItemNotSupported(Exception):
  pass
class ownershipChangeAcrossDomainNotPermitted(Exception):
  pass
class participantIsNeitherOrganizerNorAttendee(Exception):
  pass
class permissionDenied(Exception):
  pass
class permissionNotFound(Exception):
  pass
class photoNotFound(Exception):
  pass
class publishOutNotPermitted(Exception):
  pass
class queryRequiresAdminCredentials(Exception):
  pass
class quotaExceeded(Exception):
  pass
class rateLimitExceeded(Exception):
  pass
class required(Exception):
  pass
class requiredAccessLevel(Exception):
  pass
class resourceExhausted(Exception):
  pass
class resourceIdNotFound(Exception):
  pass
class resourceNotFound(Exception):
  pass
class responsePreparationFailure(Exception):
  pass
class revisionDeletionNotSupported(Exception):
  pass
class revisionNotFound(Exception):
  pass
class revisionsNotSupported(Exception):
  pass
class serviceLimit(Exception):
  pass
class serviceNotAvailable(Exception):
  pass
class shareInNotPermitted(Exception):
  pass
class shareOutNotPermitted(Exception):
  pass
class shareOutNotPermittedToUser(Exception):
  pass
class shareOutWarning(Exception):
  pass
class sharingRateLimitExceeded(Exception):
  pass
class shortcutTargetInvalid(Exception):
  pass
class storageQuotaExceeded(Exception):
  pass
class systemError(Exception):
  pass
class targetUserRoleLimitedByLicenseRestriction(Exception):
  pass
class teamDriveAlreadyExists(Exception):
  pass
class teamDriveDomainUsersOnlyRestriction(Exception):
  pass
class teamDriveTeamMembersOnlyRestriction(Exception):
  pass
class teamDriveFileLimitExceeded(Exception):
  pass
class teamDriveHierarchyTooDeep(Exception):
  pass
class teamDriveMembershipRequired(Exception):
  pass
class teamDrivesFolderMoveInNotSupported(Exception):
  pass
class teamDrivesFolderSharingNotSupported(Exception):
  pass
class teamDrivesParentLimit(Exception):
  pass
class teamDrivesSharingRestrictionNotAllowed(Exception):
  pass
class teamDrivesShortcutFileNotSupported(Exception):
  pass
class timeRangeEmpty(Exception):
  pass
class transientError(Exception):
  pass
class unknownError(Exception):
  pass
class unsupportedLanguageCode(Exception):
  pass
class unsupportedSupervisedAccount(Exception):
  pass
class uploadTooLarge(Exception):
  pass
class userCannotCreateTeamDrives(Exception):
  pass
class userAccess(Exception):
  pass
class userNotFound(Exception):
  pass
class userRateLimitExceeded(Exception):
  pass

REASON_EXCEPTION_MAP = {
  ABORTED: aborted,
  ABUSIVE_CONTENT_RESTRICTION: abusiveContentRestriction,
  ACCESS_NOT_CONFIGURED: accessNotConfigured,
  ALREADY_EXISTS: alreadyExists,
  APPLY_LABEL_FORBIDDEN: applyLabelForbidden,
  AUTH_ERROR: authError,
  BACKEND_ERROR: backendError,
  BAD_REQUEST: badRequest,
  CANNOT_ADD_PARENT: cannotAddParent,
  CANNOT_CHANGE_ORGANIZER: cannotChangeOrganizer,
  CANNOT_CHANGE_ORGANIZER_OF_INSTANCE: cannotChangeOrganizerOfInstance,
  CANNOT_CHANGE_OWN_ACL: cannotChangeOwnAcl,
  CANNOT_CHANGE_OWNER_ACL: cannotChangeOwnerAcl,
  CANNOT_CHANGE_OWN_PRIMARY_SUBSCRIPTION: cannotChangeOwnPrimarySubscription,
  CANNOT_COPY_FILE: cannotCopyFile,
  CANNOT_DELETE_ONLY_REVISION: cannotDeleteOnlyRevision,
  CANNOT_DELETE_PERMISSION: cannotDeletePermission,
  CANNOT_DELETE_PRIMARY_CALENDAR: cannotDeletePrimaryCalendar,
  CANNOT_DELETE_PRIMARY_SENDAS: cannotDeletePrimarySendAs,
  CANNOT_DELETE_RESOURCE_WITH_CHILDREN: cannotDeleteResourceWithChildren,
  CANNOT_MODIFY_INHERITED_TEAMDRIVE_PERMISSION: cannotModifyInheritedTeamDrivePermission,
  CANNOT_MODIFY_RESTRICTED_LABEL: cannotModifyRestrictedLabel,
  CANNOT_MODIFY_VIEWERS_CAN_COPY_CONTENT: cannotModifyViewersCanCopyContent,
  CANNOT_MOVE_TRASHED_ITEM_INTO_TEAMDRIVE: cannotMoveTrashedItemIntoTeamDrive,
  CANNOT_MOVE_TRASHED_ITEM_OUT_OF_TEAMDRIVE: cannotMoveTrashedItemOutOfTeamDrive,
  CANNOT_REMOVE_OWNER: cannotRemoveOwner,
  CANNOT_SET_EXPIRATION: cannotSetExpiration,
  CANNOT_SET_EXPIRATION_ON_ANYONE_OR_DOMAIN: cannotSetExpirationOnAnyoneOrDomain,
  CANNOT_SHARE_GROUPS_WITHLINK: cannotShareGroupsWithLink,
  CANNOT_SHARE_USERS_WITHLINK: cannotShareUsersWithLink,
  CANNOT_SHARE_TEAMDRIVE_TOPFOLDER_WITH_ANYONEORDOMAINS: cannotShareTeamDriveTopFolderWithAnyoneOrDomains,
  CANNOT_SHARE_TEAMDRIVE_WITH_NONGOOGLE_ACCOUNTS: cannotShareTeamDriveWithNonGoogleAccounts,
  CANNOT_UPDATE_PERMISSION: cannotUpdatePermission,
  CONDITION_NOT_MET: conditionNotMet,
  CONFLICT: conflict,
  CONTENT_OWNER_ACCOUNT_NOT_FOUND: contentOwnerAccountNotFound,
  CROSS_DOMAIN_MOVE_RESTRICTION: crossDomainMoveRestriction,
  CUSTOMER_EXCEEDED_ROLE_ASSIGNMENTS_LIMIT: customerExceededRoleAssignmentsLimit,
  CUSTOMER_NOT_FOUND: customerNotFound,
  CYCLIC_MEMBERSHIPS_NOT_ALLOWED: cyclicMembershipsNotAllowed,
  DELETED: deleted,
  DELETED_USER_NOT_FOUND: deletedUserNotFound,
  DOMAIN_ALIAS_NOT_FOUND: domainAliasNotFound,
  DOMAIN_CANNOT_USE_APIS: domainCannotUseApis,
  DOMAIN_NOT_FOUND: domainNotFound,
  DOMAIN_NOT_VERIFIED_SECONDARY: domainNotVerifiedSecondary,
  DOMAIN_POLICY: domainPolicy,
  DOWNLOAD_QUOTA_EXCEEDED: downloadQuotaExceeded,
  DUPLICATE: duplicate,
  EVENT_DURATION_EXCEEDS_LIMIT: eventDurationExceedsLimit,
  EVENT_TYPE_RESTRICTION: eventTypeRestriction,
  EXPIRATION_DATES_MUST_BE_IN_THE_FUTURE: expirationDatesMustBeInTheFuture,
  EXPIRATION_DATE_NOT_ALLOWED_FOR_SHARED_DRIVE_MEMBERS: expirationDateNotAllowedForSharedDriveMembers,
  FAILED_PRECONDITION: failedPrecondition,
  FIELD_IN_USE: fieldInUse,
  FIELD_NOT_WRITABLE: fieldNotWritable,
  FILE_NEVER_WRITABLE: fileNeverWritable,
  FILE_NOT_FOUND: fileNotFound,
  FILE_ORGANIZER_NOT_YET_ENABLED_FOR_THIS_TEAMDRIVE: fileOrganizerNotYetEnabledForThisTeamDrive,
  FILE_ORGANIZER_ON_FOLDERS_IN_SHARED_DRIVE_ONLY: fileOrganizerOnFoldersInSharedDriveOnly,
  FILE_ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED: fileOrganizerOnNonTeamDriveNotSupported,
  FILE_OWNER_NOT_MEMBER_OF_TEAMDRIVE: fileOwnerNotMemberOfTeamDrive,
  FILE_OWNER_NOT_MEMBER_OF_WRITER_DOMAIN: fileOwnerNotMemberOfWriterDomain,
  FILE_WRITER_TEAMDRIVE_MOVE_IN_DISABLED: fileWriterTeamDriveMoveInDisabled,
  FORBIDDEN: forbidden,
  GROUP_NOT_FOUND: groupNotFound,
  ILLEGAL_ACCESS_ROLE_FOR_DEFAULT: illegalAccessRoleForDefault,
  INSUFFICIENT_ADMINISTRATOR_PRIVILEGES: insufficientAdministratorPrivileges,
  INSUFFICIENT_ARCHIVED_USER_LICENSES: insufficientArchivedUserLicenses,
  INSUFFICIENT_FILE_PERMISSIONS: insufficientFilePermissions,
  INSUFFICIENT_PARENT_PERMISSIONS: insufficientParentPermissions,
  INSUFFICIENT_PERMISSIONS: insufficientPermissions,
  INTERNAL_ERROR: internalError,
  INVALID: invalid,
  INVALID_ARGUMENT: invalidArgument,
  INVALID_ATTRIBUTE_VALUE: invalidAttributeValue,
  INVALID_CUSTOMER_ID: invalidCustomerId,
  INVALID_INPUT: invalidInput,
  INVALID_LINK_VISIBILITY: invalidLinkVisibility,
  INVALID_MEMBER: invalidMember,
  INVALID_MESSAGE_ID: invalidMessageId,
  INVALID_ORGUNIT: invalidOrgunit,
  INVALID_ORGUNIT_NAME: invalidOrgunitName,
  INVALID_OWNERSHIP_TRANSFER: invalidOwnershipTransfer,
  INVALID_PARAMETER: invalidParameter,
  INVALID_PARENT_ORGUNIT: invalidParentOrgunit,
  INVALID_QUERY: invalidQuery,
  INVALID_RESOURCE: invalidResource,
  INVALID_SCHEMA_VALUE: invalidSchemaValue,
  INVALID_SCOPE_VALUE: invalidScopeValue,
  INVALID_SHARING_REQUEST: invalidSharingRequest,
  LABEL_MULTIPLE_VALUES_FOR_SINGULAR_FIELD: labelMultipleValuesForSingularField,
  LABEL_MUTATION_FORBIDDEN: labelMutationForbidden,
  LABEL_MUTATION_ILLEGAL_SELECTION: labelMutationIllegalSelection,
  LABEL_MUTATION_UNKNOWN_FIELD: labelMutationUnknownField,
  LIMIT_EXCEEDED: limitExceeded,
  LOGIN_REQUIRED: loginRequired,
  MALFORMED_WORKING_LOCATION_EVENT: malformedWorkingLocationEvent,
  MEMBER_NOT_FOUND: memberNotFound,
  NO_LIST_TEAMDRIVES_ADMINISTRATOR_PRIVILEGE: noListTeamDrivesAdministratorPrivilege,
  NO_MANAGE_TEAMDRIVE_ADMINISTRATOR_PRIVILEGE: noManageTeamDriveAdministratorPrivilege,
  NOT_A_CALENDAR_USER: notACalendarUser,
  NOT_FOUND: notFound,
  NOT_IMPLEMENTED: notImplemented,
  OPERATION_NOT_SUPPORTED: operationNotSupported,
  ORGANIZER_ON_NON_TEAMDRIVE_NOT_SUPPORTED: organizerOnNonTeamDriveNotSupported,
  ORGANIZER_ON_NON_TEAMDRIVE_ITEM_NOT_SUPPORTED: organizerOnNonTeamDriveItemNotSupported,
  ORGUNIT_NOT_FOUND: orgunitNotFound,
  OWNER_ON_TEAMDRIVE_ITEM_NOT_SUPPORTED: ownerOnTeamDriveItemNotSupported,
  OWNERSHIP_CHANGE_ACROSS_DOMAIN_NOT_PERMITTED: ownershipChangeAcrossDomainNotPermitted,
  PARTICIPANT_IS_NEITHER_ORGANIZER_NOR_ATTENDEE: participantIsNeitherOrganizerNorAttendee,
  PERMISSION_DENIED: permissionDenied,
  PERMISSION_NOT_FOUND: permissionNotFound,
  PHOTO_NOT_FOUND: photoNotFound,
  PUBLISH_OUT_NOT_PERMITTED: publishOutNotPermitted,
  QUERY_REQUIRES_ADMIN_CREDENTIALS: queryRequiresAdminCredentials,
  QUOTA_EXCEEDED: quotaExceeded,
  RATE_LIMIT_EXCEEDED: rateLimitExceeded,
  REQUIRED: required,
  REQUIRED_ACCESS_LEVEL: requiredAccessLevel,
  RESOURCE_EXHAUSTED: resourceExhausted,
  RESOURCE_ID_NOT_FOUND: resourceIdNotFound,
  RESOURCE_NOT_FOUND: resourceNotFound,
  RESPONSE_PREPARATION_FAILURE: responsePreparationFailure,
  REVISION_DELETION_NOT_SUPPORTED: revisionDeletionNotSupported,
  REVISION_NOT_FOUND: revisionNotFound,
  REVISIONS_NOT_SUPPORTED: revisionsNotSupported,
  SERVICE_LIMIT: serviceLimit,
  SERVICE_NOT_AVAILABLE: serviceNotAvailable,
  SHARE_IN_NOT_PERMITTED: shareInNotPermitted,
  SHARE_OUT_NOT_PERMITTED: shareOutNotPermitted,
  SHARE_OUT_NOT_PERMITTED_TO_USER: shareOutNotPermittedToUser,
  SHARE_OUT_WARNING: shareOutWarning,
  SHARING_RATE_LIMIT_EXCEEDED: sharingRateLimitExceeded,
  SHORTCUT_TARGET_INVALID: shortcutTargetInvalid,
  STORAGE_QUOTA_EXCEEDED: storageQuotaExceeded,
  SYSTEM_ERROR: systemError,
  TARGET_USER_ROLE_LIMITED_BY_LICENSE_RESTRICTION: targetUserRoleLimitedByLicenseRestriction,
  TEAMDRIVE_ALREADY_EXISTS: teamDriveAlreadyExists,
  TEAMDRIVE_DOMAIN_USERS_ONLY_RESTRICTION: teamDriveDomainUsersOnlyRestriction,
  TEAMDRIVE_TEAM_MEMBERS_ONLY_RESTRICTION: teamDriveTeamMembersOnlyRestriction,
  TEAMDRIVE_FILE_LIMIT_EXCEEDED: teamDriveFileLimitExceeded,
  TEAMDRIVE_HIERARCHY_TOO_DEEP: teamDriveHierarchyTooDeep,
  TEAMDRIVE_MEMBERSHIP_REQUIRED: teamDriveMembershipRequired,
  TEAMDRIVES_FOLDER_MOVE_IN_NOT_SUPPORTED: teamDrivesFolderMoveInNotSupported,
  TEAMDRIVES_FOLDER_SHARING_NOT_SUPPORTED: teamDrivesFolderSharingNotSupported,
  TEAMDRIVES_PARENT_LIMIT: teamDrivesParentLimit,
  TEAMDRIVES_SHARING_RESTRICTION_NOT_ALLOWED: teamDrivesSharingRestrictionNotAllowed,
  TEAMDRIVES_SHORTCUT_FILE_NOT_SUPPORTED: teamDrivesShortcutFileNotSupported,
  TIME_RANGE_EMPTY: timeRangeEmpty,
  TRANSIENT_ERROR: transientError,
  UNKNOWN_ERROR: unknownError,
  UNSUPPORTED_LANGUAGE_CODE: unsupportedLanguageCode,
  UNSUPPORTED_SUPERVISED_ACCOUNT: unsupportedSupervisedAccount,
  UPLOAD_TOO_LARGE: uploadTooLarge,
  USER_CANNOT_CREATE_TEAMDRIVES: userCannotCreateTeamDrives,
  USER_ACCESS: userAccess,
  USER_NOT_FOUND: userNotFound,
  USER_RATE_LIMIT_EXCEEDED: userRateLimitExceeded,
  }



#!/usr/bin/env python
# Copyright (C) 2008 Google, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Allow Google Apps domain administrators to audit user data.

  AuditService: Set auditing."""

__author__ = 'jlee@pbu.edu'

from base64 import b64encode

import gdata.apps
import gdata.apps.service
import gdata.service

class AuditService(gdata.apps.service.PropertyService):
  """Client for the Google Apps Audit service."""

  def _serviceUrl(self, setting_id, domain=None, user=None):
    if domain is None:
      domain = self.domain
    if user is None:
      return '/a/feeds/compliance/audit/%s/%s' % (setting_id, domain)
    else:
      return '/a/feeds/compliance/audit/%s/%s/%s' % (setting_id, domain, user)

  def updatePGPKey(self, pgpkey):
    """Updates Public PGP Key Google uses to encrypt audit data

    Args:
      pgpkey: string, ASCII text of PGP Public Key to be used

    Returns:
      A dict containing the result of the POST operation."""

    uri = self._serviceUrl('publickey')
    b64pgpkey = b64encode(pgpkey)
    properties = {}
    properties['publicKey'] = b64pgpkey
    return self._PostProperties(uri, properties)

  def createEmailMonitor(self, source_user, destination_user, end_date, 
                         begin_date=None, incoming_headers_only=False, 
                         outgoing_headers_only=False, drafts=False, 
                         drafts_headers_only=False, chats=False, 
                         chats_headers_only=False):
    """Creates a email monitor, forwarding the source_users emails/chats

    Args:
      source_user: string, the user whose email will be audited
      destination_user: string, the user to receive the audited email
      end_date: string, the date the audit will end in
                "yyyy-MM-dd HH:mm" format, required
      begin_date: string, the date the audit will start in 
                  "yyyy-MM-dd HH:mm" format, leave blank to use current time
      incoming_headers_only: boolean, whether to audit only the headers of
                             mail delivered to source user
      outgoing_headers_only: boolean, whether to audit only the headers of
                             mail sent from the source user
      drafts: boolean, whether to audit draft messages of the source user
      drafts_headers_only: boolean, whether to audit only the headers of
                           mail drafts saved by the user
      chats: boolean, whether to audit archived chats of the source user
      chats_headers_only: boolean, whether to audit only the headers of
                          archived chats of the source user

    Returns:
      A dict containing the result of the POST operation."""

    uri = self._serviceUrl('mail/monitor', user=source_user)
    properties = {}
    properties['destUserName'] = destination_user
    if begin_date is not None:
      properties['beginDate'] = begin_date
    properties['endDate'] = end_date
    if incoming_headers_only:
      properties['incomingEmailMonitorLevel'] = 'HEADER_ONLY'
    else:
      properties['incomingEmailMonitorLevel'] = 'FULL_MESSAGE'
    if outgoing_headers_only:
      properties['outgoingEmailMonitorLevel'] = 'HEADER_ONLY'
    else:
      properties['outgoingEmailMonitorLevel'] = 'FULL_MESSAGE'
    if drafts:
      if drafts_headers_only:
        properties['draftMonitorLevel'] = 'HEADER_ONLY'
      else:
        properties['draftMonitorLevel'] = 'FULL_MESSAGE'
    if chats:
      if chats_headers_only:
        properties['chatMonitorLevel'] = 'HEADER_ONLY'
      else:
        properties['chatMonitorLevel'] = 'FULL_MESSAGE'
    return self._PostProperties(uri, properties)

  def getEmailMonitors(self, user):
    """"Gets the email monitors for the given user

    Args:
      user: string, the user to retrieve email monitors for

    Returns:
      list results of the POST operation

    """
    uri = self._serviceUrl('mail/monitor', user=user)
    return self._GetPropertiesList(uri)

  def deleteEmailMonitor(self, source_user, destination_user):
    """Deletes the email monitor for the given user

    Args:
      source_user: string, the user who is being monitored
      destination_user: string, theuser who recieves the monitored emails

    Returns:
      Nothing
    """

    uri = self._serviceUrl('mail/monitor', user=source_user+'/'+destination_user)
    try:
      return self._DeleteProperties(uri)
    except gdata.service.RequestError as e:
      raise gdata.apps.service.AppsForYourDomainException(e.args[0])

  def createAccountInformationRequest(self, user):
    """Creates a request for account auditing details

    Args:
      user: string, the user to request account information for

    Returns:
      A dict containing the result of the post operation."""

    uri = self._serviceUrl('account', user=user)
    properties = {}
    #XML Body is left empty
    try:
      return self._PostProperties(uri, properties)
    except gdata.service.RequestError as e:
      raise gdata.apps.service.AppsForYourDomainException(e.args[0])

  def getAccountInformationRequestStatus(self, user, request_id):
    """Gets the status of an account auditing request

    Args:
      user: string, the user whose account auditing details were requested
      request_id: string, the request_id

    Returns:
      A dict containing the result of the get operation."""

    uri = self._serviceUrl('account', user=user+'/'+request_id)
    try:
      return self._GetProperties(uri)
    except gdata.service.RequestError as e:
      raise gdata.apps.service.AppsForYourDomainException(e.args[0])

  def getAllAccountInformationRequestsStatus(self):
    """Gets the status of all account auditing requests for the domain

    Args:
      None

    Returns:
      list results of the POST operation
    """

    uri = self._serviceUrl('account')
    return self._GetPropertiesList(uri)


  def deleteAccountInformationRequest(self, user, request_id):
    """Deletes the request for account auditing information

   Args:
     user: string, the user whose account auditing details were requested
     request_id: string, the request_id

   Returns:
     Nothing
   """

    uri = self._serviceUrl('account', user=user+'/'+request_id)
    try:
      return self._DeleteProperties(uri)
    except gdata.service.RequestError as e:
      raise gdata.apps.service.AppsForYourDomainException(e.args[0])

  def createMailboxExportRequest(self, user, begin_date=None, end_date=None, include_deleted=False, search_query=None, headers_only=False):
    """Creates a mailbox export request

    Args:
      user: string, the user whose mailbox export is being requested
      begin_date: string, date of earliest emails to export, optional, defaults to date of account creation
                  format is 'yyyy-MM-dd HH:mm'
      end_date: string, date of latest emails to export, optional, defaults to current date
                format is 'yyyy-MM-dd HH:mm'
      include_deleted: boolean, whether to include deleted emails in export, mutually exclusive with search_query
      search_query: string, gmail style search query, matched emails will be exported, mutually exclusive with include_deleted

    Returns:
      A dict containing the result of the post operation."""

    uri = self._serviceUrl('mail/export', user=user)
    properties = {}
    if begin_date is not None:
      properties['beginDate'] = begin_date
    if end_date is not None:
      properties['endDate'] = end_date
    if include_deleted is not None:
      properties['includeDeleted'] = gdata.apps.service._bool2str(include_deleted)
    if search_query is not None:
      properties['searchQuery'] = search_query
    if headers_only is True:
      properties['packageContent'] = 'HEADER_ONLY'
    else:
      properties['packageContent'] = 'FULL_MESSAGE'
    return self._PostProperties(uri, properties)

  def getMailboxExportRequestStatus(self, user, request_id):
    """Gets the status of an mailbox export request

    Args:
      user: string, the user whose mailbox were requested
      request_id: string, the request_id

    Returns:
      A dict containing the result of the get operation."""

    uri = self._serviceUrl('mail/export', user=user+'/'+request_id)
    try:
      return self._GetProperties(uri)
    except gdata.service.RequestError as e:
      raise gdata.apps.service.AppsForYourDomainException(e.args[0])

  def getAllMailboxExportRequestsStatus(self):
    """Gets the status of all mailbox export requests for the domain

    Args:
      None

    Returns:
      list results of the POST operation
    """

    uri = self._serviceUrl('mail/export')
    return self._GetPropertiesList(uri)


  def deleteMailboxExportRequest(self, user, request_id):
    """Deletes the request for mailbox export

   Args:
     user: string, the user whose mailbox were requested
     request_id: string, the request_id

   Returns:
     Nothing
   """

    uri = self._serviceUrl('mail/export', user=user+'/'+request_id)
    try:
      return self._DeleteProperties(uri)
    except gdata.service.RequestError as e:
      raise gdata.apps.service.AppsForYourDomainException(e.args[0])



#!/usr/bin/env python
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License 2.0;



# This module is used for version 2 of the Google Data APIs.


# __author__ = 'j.s@google.com (Jeff Scudder)'

import io
import os.path
import pickle
import tempfile

import atom.http_core


class Error(Exception):
    pass


class NoRecordingFound(Error):
    pass


class MockHttpClient(object):
    debug = None
    real_client = None
    last_request_was_live = False

    # The following members are used to construct the session cache temp file
    # name.
    # These are combined to form the file name
    # /tmp/cache_prefix.cache_case_name.cache_test_name
    cache_name_prefix = 'gdata_live_test'
    cache_case_name = ''
    cache_test_name = ''

    def __init__(self, recordings=None, real_client=None):
        self._recordings = recordings or []
        if real_client is not None:
            self.real_client = real_client

    def add_response(self, http_request, status, reason, headers=None,
                     body=None):
        response = MockHttpResponse(status, reason, headers, body)
        # TODO Scrub the request and the response.
        self._recordings.append((http_request._copy(), response))

    AddResponse = add_response

    def request(self, http_request):
        """Provide a recorded response, or record a response for replay.

        If the real_client is set, the request will be made using the
        real_client, and the response from the server will be recorded.
        If the real_client is None (the default), this method will examine
        the recordings and find the first which matches.
        """
        request = http_request._copy()
        _scrub_request(request)
        if self.real_client is None:
            self.last_request_was_live = False
            for recording in self._recordings:
                if _match_request(recording[0], request):
                    return recording[1]
        else:
            # Pass along the debug settings to the real client.
            self.real_client.debug = self.debug
            # Make an actual request since we can use the real HTTP client.
            self.last_request_was_live = True
            response = self.real_client.request(http_request)
            scrubbed_response = _scrub_response(response)
            self.add_response(request, scrubbed_response.status,
                              scrubbed_response.reason,
                              dict(atom.http_core.get_headers(scrubbed_response)),
                              scrubbed_response.read())
            # Return the recording which we just added.
            return self._recordings[-1][1]
        raise NoRecordingFound('No recoding was found for request: %s %s' % (
            request.method, str(request.uri)))

    Request = request

    def _save_recordings(self, filename):
        recording_file = open(os.path.join(tempfile.gettempdir(), filename),
                              'wb')
        pickle.dump(self._recordings, recording_file)
        recording_file.close()

    def _load_recordings(self, filename):
        recording_file = open(os.path.join(tempfile.gettempdir(), filename),
                              'rb')
        self._recordings = pickle.load(recording_file)
        recording_file.close()

    def _delete_recordings(self, filename):
        full_path = os.path.join(tempfile.gettempdir(), filename)
        if os.path.exists(full_path):
            os.remove(full_path)

    def _load_or_use_client(self, filename, http_client):
        if os.path.exists(os.path.join(tempfile.gettempdir(), filename)):
            self._load_recordings(filename)
        else:
            self.real_client = http_client

    def use_cached_session(self, name=None, real_http_client=None):
        """Attempts to load recordings from a previous live request.

        If a temp file with the recordings exists, then it is used to fulfill
        requests. If the file does not exist, then a real client is used to
        actually make the desired HTTP requests. Requests and responses are
        recorded and will be written to the desired temprary cache file when
        close_session is called.

        Args:
          name: str (optional) The file name of session file to be used. The file
                is loaded from the temporary directory of this machine. If no name
                is passed in, a default name will be constructed using the
                cache_name_prefix, cache_case_name, and cache_test_name of this
                object.
          real_http_client: atom.http_core.HttpClient the real client to be used
                            if the cached recordings are not found. If the default
                            value is used, this will be an
                            atom.http_core.HttpClient.
        """
        if real_http_client is None:
            real_http_client = atom.http_core.HttpClient()
        if name is None:
            self._recordings_cache_name = self.get_cache_file_name()
        else:
            self._recordings_cache_name = name
        self._load_or_use_client(self._recordings_cache_name, real_http_client)

    def close_session(self):
        """Saves recordings in the temporary file named in use_cached_session."""
        if self.real_client is not None:
            self._save_recordings(self._recordings_cache_name)

    def delete_session(self, name=None):
        """Removes recordings from a previous live request."""
        if name is None:
            self._delete_recordings(self._recordings_cache_name)
        else:
            self._delete_recordings(name)

    def get_cache_file_name(self):
        return '%s.%s.%s' % (self.cache_name_prefix, self.cache_case_name,
                             self.cache_test_name)

    def _dump(self):
        """Provides debug information in a string."""
        output = 'MockHttpClient\n  real_client: %s\n  cache file name: %s\n' % (
            self.real_client, self.get_cache_file_name())
        output += '  recordings:\n'
        i = 0
        for recording in self._recordings:
            output += '    recording %i is for: %s %s\n' % (
                i, recording[0].method, str(recording[0].uri))
            i += 1
        return output


def _match_request(http_request, stored_request):
    """Determines whether a request is similar enough to a stored request
       to cause the stored response to be returned."""
    # Check to see if the host names match.
    if (http_request.uri.host is not None
        and http_request.uri.host != stored_request.uri.host):
        return False
    # Check the request path in the URL (/feeds/private/full/x)
    elif http_request.uri.path != stored_request.uri.path:
        return False
    # Check the method used in the request (GET, POST, etc.)
    elif http_request.method != stored_request.method:
        return False
    # If there is a gsession ID in either request, make sure that it is matched
    # exactly.
    elif ('gsessionid' in http_request.uri.query
          or 'gsessionid' in stored_request.uri.query):
        if 'gsessionid' not in stored_request.uri.query:
            return False
        elif 'gsessionid' not in http_request.uri.query:
            return False
        elif (http_request.uri.query['gsessionid']
                  != stored_request.uri.query['gsessionid']):
            return False
    # Ignores differences in the query params (?start-index=5&max-results=20),
    # the body of the request, the port number, HTTP headers, just to name a
    # few.
    return True


def _scrub_request(http_request):
    """ Removes email address and password from a client login request.

    Since the mock server saves the request and response in plantext, sensitive
    information like the password should be removed before saving the
    recordings. At the moment only requests sent to a ClientLogin url are
    scrubbed.
    """
    if (http_request and http_request.uri and http_request.uri.path and
            http_request.uri.path.endswith('ClientLogin')):
        # Remove the email and password from a ClientLogin request.
        http_request._body_parts = []
        http_request.add_form_inputs(
            {'form_data': 'client login request has been scrubbed'})
    else:
        # We can remove the body of the post from the recorded request, since
        # the request body is not used when finding a matching recording.
        http_request._body_parts = []
    return http_request


def _scrub_response(http_response):
    return http_response


class EchoHttpClient(object):
    """Sends the request data back in the response.

    Used to check the formatting of the request as it was sent. Always responds
    with a 200 OK, and some information from the HTTP request is returned in
    special Echo-X headers in the response. The following headers are added
    in the response:
    'Echo-Host': The host name and port number to which the HTTP connection is
                 made. If no port was passed in, the header will contain
                 host:None.
    'Echo-Uri': The path portion of the URL being requested. /example?x=1&y=2
    'Echo-Scheme': The beginning of the URL, usually 'http' or 'https'
    'Echo-Method': The HTTP method being used, 'GET', 'POST', 'PUT', etc.
    """

    def request(self, http_request):
        return self._http_request(http_request.uri, http_request.method,
                                  http_request.headers, http_request._body_parts)

    def _http_request(self, uri, method, headers=None, body_parts=None):
        body = io.StringIO()
        response = atom.http_core.HttpResponse(status=200, reason='OK', body=body)
        if headers is None:
            response._headers = {}
        else:
            # Copy headers from the request to the response but convert values to
            # strings. Server response headers always come in as strings, so an int
            # should be converted to a corresponding string when echoing.
            for header, value in headers.items():
                response._headers[header] = str(value)
        response._headers['Echo-Host'] = '%s:%s' % (uri.host, str(uri.port))
        response._headers['Echo-Uri'] = uri._get_relative_path()
        response._headers['Echo-Scheme'] = uri.scheme
        response._headers['Echo-Method'] = method
        for part in body_parts:
            if isinstance(part, str):
                body.write(part)
            elif hasattr(part, 'read'):
                body.write(part.read())
        body.seek(0)
        return response


class SettableHttpClient(object):
    """An HTTP Client which responds with the data given in set_response."""

    def __init__(self, status, reason, body, headers):
        """Configures the response for the server.

        See set_response for details on the arguments to the constructor.
        """
        self.set_response(status, reason, body, headers)
        self.last_request = None

    def set_response(self, status, reason, body, headers):
        """Determines the response which will be sent for each request.

        Args:
          status: An int for the HTTP status code, example: 200, 404, etc.
          reason: String for the HTTP reason, example: OK, NOT FOUND, etc.
          body: The body of the HTTP response as a string or a file-like
                object (something with a read method).
          headers: dict of strings containing the HTTP headers in the response.
        """
        self.response = atom.http_core.HttpResponse(status=status, reason=reason,
                                                    body=body)
        self.response._headers = headers.copy()

    def request(self, http_request):
        self.last_request = http_request
        return self.response


class MockHttpResponse(atom.http_core.HttpResponse):
    def __init__(self, status=None, reason=None, headers=None, body=None):
        self._headers = headers or {}
        if status is not None:
            self.status = status
        if reason is not None:
            self.reason = reason
        if body is not None:
            # Instead of using a file-like object for the body, store as a string
            # so that reads can be repeated.
            if hasattr(body, 'read'):
                self._body = body.read()
            else:
                self._body = body

    def read(self):
        return self._body



# -*- coding: utf-8 -*-

# Copyright (C) 2025 Ross Scroggs All Rights Reserved.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GAM global variables

"""

# The following GM_XXX constants are arbitrary but must be unique
# Most errors print a message and bail out with a return code
# Some commands want to set a non-zero return code but not bail
# GAM admin user from oauth2.txt or oauth2service.json
ADMIN = 'admn'
# Number/length of API call retries
API_CALLS_RETRY_DATA = 'rtry'
# GAM cache directory. If no_cache is True, this variable will be set to None
CACHE_DIR = 'gacd'
# Reset GAM cache directory after discovery
CACHE_DISCOVERY_ONLY = 'gcdo'
# Classroom service not available
CLASSROOM_SERVICE_NOT_AVAILABLE = 'csna'
# Command logging
CMDLOG_HANDLER = 'clha'
CMDLOG_LOGGER = 'cllo'
# Convert to local time
CONVERT_TO_LOCAL_TIME = 'ctlt'
# Credentials scopes
CREDENTIALS_SCOPES = 'crsc'
# csvfile keyfield <FieldName> [delimiter <Character>] (matchfield <FieldName> <MatchPattern>)* [datafield <FieldName>(:<FieldName>*) [delimiter <String>]]
CSVFILE = 'csvf'
# { key: [datafieldvalues]}
CSV_DATA_DICT = 'csdd'
CSV_KEY_FIELD = 'cskf'
CSV_SUBKEY_FIELD = 'cssk'
CSV_DATA_FIELD = 'csdf'
# Filter for input column drop values
CSV_INPUT_ROW_DROP_FILTER = 'cird'
# Mode (and|or) for input column drop values
CSV_INPUT_ROW_DROP_FILTER_MODE = 'cidm'
# Filter for input column values
CSV_INPUT_ROW_FILTER = 'cirf'
# Mode (and|or) for input column values
CSV_INPUT_ROW_FILTER_MODE = 'cirm'
# Limit number of input rows
CSV_INPUT_ROW_LIMIT = 'cirl'
# Column delimiter in CSV output file
CSV_OUTPUT_COLUMN_DELIMITER = 'codl'
# Filter for output column headers to drop
CSV_OUTPUT_HEADER_DROP_FILTER = 'cohd'
# Filter for output column headers
CSV_OUTPUT_HEADER_FILTER = 'cohf'
# Force output column headers
CSV_OUTPUT_HEADER_FORCE = 'cofh'
# Order output column headers
CSV_OUTPUT_HEADER_ORDER = 'coho'
# No escape character in CSV output file
CSV_OUTPUT_NO_ESCAPE_CHAR = 'cone'
# Quote character in CSV output file
CSV_OUTPUT_QUOTE_CHAR = 'coqc'
# Filter for output column drop values
CSV_OUTPUT_ROW_DROP_FILTER = 'cord'
# Mode (and|or) for output column drop values
CSV_OUTPUT_ROW_DROP_FILTER_MODE = 'codm'
# Filter for output column values
CSV_OUTPUT_ROW_FILTER = 'corf'
# Mode (and|or) for output column values
CSV_OUTPUT_ROW_FILTER_MODE = 'corm'
# Limit number of output rows
CSV_OUTPUT_ROW_LIMIT = 'corl'
# Add timestamp column to CSV output file
CSV_OUTPUT_TIMESTAMP_COLUMN = 'cotc'
# Transpose output rows/columns
CSV_OUTPUT_TRANSPOSE = 'cotr'
# Output sort headers
CSV_OUTPUT_SORT_HEADERS = 'cosh'
# CSV todrive options
CSV_TODRIVE = 'todr'
# Current API services
CURRENT_API_SERVICES = 'caps'
# Current Client API
CURRENT_CLIENT_API = 'ccap'
# Current Client API scopes
CURRENT_CLIENT_API_SCOPES = 'ccas'
# Current Service Account API
CURRENT_SVCACCT_API = 'csap'
# Current Service Account API scopes
CURRENT_SVCACCT_API_SCOPES = 'csas'
# Current Service Account user
CURRENT_SVCACCT_USER = 'csa'
# datetime.datetime.now
DATETIME_NOW = 'dtno'
# If debug_level > 0: extra_args['prettyPrint'] = True, httplib2.debuglevel = gam_debug_level, appsObj.debug = True
DEBUG_LEVEL = 'dbgl'
# Decoded ID token
DECODED_ID_TOKEN = 'didt'
# Index of start of <UserTypeEntity> in command line
ENTITY_CL_DELAY_START = 'ecld'
ENTITY_CL_START = 'ecls'
# Extra arguments to pass to GAPI functions
EXTRA_ARGS_LIST = 'exad'
# gam.cfg file
GAM_CFG_FILE = 'gcfi'
GAM_CFG_PATH = 'gcpa'
GAM_CFG_SECTION = 'gcse'
GAM_CFG_SECTION_NAME = 'gcsn'
# Path to gam
GAM_PATH = 'gpth'
# Python source, PyInstaller or StaticX?
GAM_TYPE = 'gtyp'
# Length of last Got message
LAST_GOT_MSG_LEN = 'lgml'
# License SKUs
LICENSE_SKUS = 'lsku'
# Make Building ID/Name map
MAKE_BUILDING_ID_NAME_MAP = 'mkbm'
# Dictionary mapping Building ID to Name
MAP_BUILDING_ID_TO_NAME = 'bi2n'
# Dictionary mapping Building Name to ID
MAP_BUILDING_NAME_TO_ID = 'bn2i'
# Dictionary mapping OrgUnit ID to Name
MAP_ORGUNIT_ID_TO_NAME = 'oi2n'
# Dictionary mapping Shared Drive ID to Name
MAP_SHAREDDRIVE_ID_TO_NAME = 'si2n'
# Make Role ID/Name map
MAKE_ROLE_ID_NAME_MAP = 'mkrm'
# Dictionary mapping Role ID to Name
MAP_ROLE_ID_TO_NAME = 'ri2n'
# Dictionary mapping Role Name to ID
MAP_ROLE_NAME_TO_ID = 'rn2i'
# Dictionary mapping User ID to Name
MAP_USER_ID_TO_NAME = 'ui2n'
# Multiprocess exit condition
MULTIPROCESS_EXIT_CONDITION = 'mpec'
# Multiprocess exit processing
MULTIPROCESS_EXIT_PROCESSING = 'mpep'
# Number of batch items
NUM_BATCH_ITEMS = 'nbat'
# Values retrieved from oauth2service.json
OAUTH2SERVICE_CLIENT_ID = 'osci'
OAUTH2SERVICE_JSON_DATA = 'osjd'
# Values retrieved from oauth2.txt
OAUTH2_CLIENT_ID = 'oaci'
# oauth2.txt lock file
OAUTH2_TXT_LOCK = 'oatl'
# Output date format, empty defalts to ISOFormat
OUTPUT_DATEFORMAT = 'oudf'
# Output time format, empty defalts to ISOFormat
OUTPUT_TIMEFORMAT = 'outf'
# gam.cfg parser
PARSER = 'pars'
# Process ID
PID = 'pid '
# Domains for print alises|groups|users
PRINT_AGU_DOMAINS = 'pagu'
# OrgUnits for print cros
PRINT_CROS_OUS = 'pcou'
# OrgUnits and children for print cros
PRINT_CROS_OUS_AND_CHILDREN = 'pcoc'
# Check API calls rate
RATE_CHECK_COUNT = 'rccn'
RATE_CHECK_START = 'rcst'
# Section name from outer gam, passed to inner gams
SECTION = 'sect'
# Enable/disable "Getting ... " messages
SHOW_GETTINGS = 'shog'
# Enable/disable NL at end of "Got ..." messages
SHOW_GETTINGS_GOT_NL = 'shgn'
# redirected files
SAVED_STDOUT = 'svso'
STDERR = 'stde'
STDOUT = 'stdo'
# Scopes values retrieved from oauth2service.json
SVCACCT_SCOPES = 'sasc'
# Were scopes values retrieved from oauth2service.json
SVCACCT_SCOPES_DEFINED = 'sasd'
# Most errors print a message and bail out with a return code
# Some commands want to set a non-zero return code but not bail
SYSEXITRC = 'sxrc'
# Encodings
SYS_ENCODING = 'syen'
# Shared by threadBatchWorker and threadBatchGAMCommands
TBATCH_QUEUE = 'batq'
# redirected file fields: name, mode, encoding, write header, multiproces, queue
REDIRECT_NAME = 'rdfn'
REDIRECT_MODE = 'rdmo'
REDIRECT_FD = 'rdfd'
REDIRECT_MULTI_FD = 'rdmf'
REDIRECT_STD = 'rdst'
REDIRECT_ENCODING = 'rden'
REDIRECT_WRITE_HEADER = 'rdwh'
REDIRECT_MULTIPROCESS = 'rdmp'
REDIRECT_QUEUE = 'rdq'
REDIRECT_QUEUE_NAME = 'name'
REDIRECT_QUEUE_CLEAR_ROW_FILTERS = 'clearRowFilters'
REDIRECT_QUEUE_TODRIVE = 'todrive'
REDIRECT_QUEUE_CSVPF = 'csvpf'
REDIRECT_QUEUE_DATA = 'rows'
REDIRECT_QUEUE_ARGS = 'args'
REDIRECT_QUEUE_GLOBALS = 'globals'
REDIRECT_QUEUE_VALUES = 'values'
REDIRECT_QUEUE_START = 'start'
REDIRECT_QUEUE_END = 'end'
REDIRECT_QUEUE_EOF = 'eof'
#
Globals = {
  ADMIN: None,
  API_CALLS_RETRY_DATA: {},
  CACHE_DIR: None,
  CACHE_DISCOVERY_ONLY: True,
  CLASSROOM_SERVICE_NOT_AVAILABLE: False,
  CMDLOG_HANDLER: None,
  CMDLOG_LOGGER: None,
  CONVERT_TO_LOCAL_TIME: False,
  CREDENTIALS_SCOPES: set(),
  CSVFILE: {},
  CSV_DATA_DICT: {},
  CSV_KEY_FIELD: None,
  CSV_SUBKEY_FIELD: None,
  CSV_DATA_FIELD: None,
  CSV_INPUT_ROW_DROP_FILTER: [],
  CSV_INPUT_ROW_DROP_FILTER_MODE: False,
  CSV_INPUT_ROW_FILTER: [],
  CSV_INPUT_ROW_FILTER_MODE: True,
  CSV_INPUT_ROW_LIMIT: 0,
  CSV_OUTPUT_COLUMN_DELIMITER: None,
  CSV_OUTPUT_HEADER_DROP_FILTER: [],
  CSV_OUTPUT_HEADER_FILTER: [],
  CSV_OUTPUT_HEADER_FORCE: [],
  CSV_OUTPUT_HEADER_ORDER: [],
  CSV_OUTPUT_NO_ESCAPE_CHAR: None,
  CSV_OUTPUT_QUOTE_CHAR: None,
  CSV_OUTPUT_ROW_DROP_FILTER: [],
  CSV_OUTPUT_ROW_DROP_FILTER_MODE: False,
  CSV_OUTPUT_ROW_FILTER: [],
  CSV_OUTPUT_ROW_FILTER_MODE: True,
  CSV_OUTPUT_ROW_LIMIT: 0,
  CSV_OUTPUT_SORT_HEADERS: [],
  CSV_OUTPUT_TIMESTAMP_COLUMN: None,
  CSV_OUTPUT_TRANSPOSE: False,
  CSV_TODRIVE: {},
  CURRENT_API_SERVICES: {},
  CURRENT_CLIENT_API: None,
  CURRENT_CLIENT_API_SCOPES: set(),
  CURRENT_SVCACCT_API: None,
  CURRENT_SVCACCT_API_SCOPES: set(),
  CURRENT_SVCACCT_USER: None,
  DATETIME_NOW: None,
  DEBUG_LEVEL: 0,
  DECODED_ID_TOKEN: None,
  ENTITY_CL_DELAY_START: 1,
  ENTITY_CL_START: 1,
  EXTRA_ARGS_LIST: [],
  GAM_CFG_FILE: '',
  GAM_CFG_PATH: '',
  GAM_CFG_SECTION: '',
  GAM_CFG_SECTION_NAME: '',
  GAM_PATH: '.',
  GAM_TYPE: '',
  LAST_GOT_MSG_LEN: 0,
  LICENSE_SKUS: [],
  MAKE_BUILDING_ID_NAME_MAP: True,
  MAKE_ROLE_ID_NAME_MAP: True,
  MAP_BUILDING_ID_TO_NAME: {},
  MAP_BUILDING_NAME_TO_ID: {},
  MAP_ORGUNIT_ID_TO_NAME: {},
  MAP_SHAREDDRIVE_ID_TO_NAME: {},
  MAP_ROLE_ID_TO_NAME: {},
  MAP_ROLE_NAME_TO_ID: {},
  MAP_USER_ID_TO_NAME: {},
  MULTIPROCESS_EXIT_CONDITION: None,
  MULTIPROCESS_EXIT_PROCESSING: False,
  NUM_BATCH_ITEMS: 0,
  OAUTH2SERVICE_CLIENT_ID: None,
  OAUTH2SERVICE_JSON_DATA: {},
  OAUTH2_CLIENT_ID: None,
  OAUTH2_TXT_LOCK: None,
  OUTPUT_DATEFORMAT: '',
  OUTPUT_TIMEFORMAT: '',
  PARSER: None,
  PID: 0,
  PRINT_AGU_DOMAINS: '',
  PRINT_CROS_OUS: '',
  PRINT_CROS_OUS_AND_CHILDREN: '',
  RATE_CHECK_COUNT: 0,
  RATE_CHECK_START: 0,
  SECTION: None,
  SHOW_GETTINGS: True,
  SHOW_GETTINGS_GOT_NL: False,
  SAVED_STDOUT: None,
  STDERR: {},
  STDOUT: {},
  SVCACCT_SCOPES: {},
  SVCACCT_SCOPES_DEFINED: False,
  SYSEXITRC: 0,
  SYS_ENCODING: 'utf-8',
  TBATCH_QUEUE: None
  }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Sample Python script to call GAM"""

import multiprocessing
import platform

from gam import initializeLogging, CallGAMCommand

if __name__ == '__main__':
# One time initialization
  if platform.system() != 'Linux':
    multiprocessing.freeze_support()
    multiprocessing.set_start_method('spawn')
  initializeLogging()
#
  CallGAMCommand(['gam', 'version'])
  # Issue command, output goes to stdout/stderr
  rc = CallGAMCommand(['gam', 'info', 'domain'])
  # Issue command, redirect stdout/stderr
  rc = CallGAMCommand(['gam', 'redirect', 'stdout', 'domain.txt', 'redirect', 'stderr', 'stdout', 'info', 'domain'])
